import numpy as np
from itertools import *
import pandas as pd
import logging
import time
from pysnptools.distreader import DistReader
from pysnptools.snpreader import SnpReader
import pysnptools.standardizer as stdizer

class _Snp2Dist(DistReader):

    def __init__(self, snpreader, max_weight=2.0, block_size=None):
        super(_Snp2Dist, self).__init__()

        self.snpreader = snpreader
        self.max_weight = max_weight
        self.block_size = block_size

    @property
    def row(self):
        return self.snpreader.iid

    @property
    def col(self):
        return self.snpreader.col

    def __repr__(self):
        return self._internal_repr()

    def _internal_repr(self): #!!! merge this with __repr__
        s = "{0}.as_dist(".format(self.snpreader)
        if self.block_size is not None:
            s += "block_size={0}".format(self.block_size)
        s += ")"
        return s

    def copyinputs(self, copier):
        #Doesn't need run_once
        copier.input(self.snpreader)

    def _snpval_to_distval(self, snpval, order, dtype):
        if order=='A':
            order='F'
        dtype = np.dtype(dtype)

        distval = np.zeros([snpval.shape[0],snpval.shape[1],3],dtype=dtype,order=order)
        distval = distval.reshape(-1,distval.shape[-1],order=order)
        factor = 2.0/self.max_weight
        for count in range(distval.shape[-1]):
            bool_array = snpval.reshape(-1,order=order)*factor==count
            distval[bool_array,count]=1
        distval[distval.sum(axis=-1)!=1,:]=np.nan #replace any dist that doesn't sum to 1 with nan
        distval = distval.reshape([snpval.shape[0],snpval.shape[1],distval.shape[-1]],order=order)
        return distval

    def _read(self, row_index_or_none, col_index_or_none, order, dtype, force_python_only, view_ok, num_threads):
        assert row_index_or_none is None and col_index_or_none is None #real assert because indexing should already be pushed to the inner snpreader
        dtype = np.dtype(dtype)

        #Do all-at-once (not in blocks) if 1. No block size is given or 2. The #ofSNPs < Min(block_size,iid_count)
        if self.block_size is None or (self.sid_count <= self.block_size or self.sid_count <= self.iid_count):
            snpdata,_ = SnpReader._as_snpdata(self.snpreader,dtype=dtype,order=order,force_python_only=force_python_only,standardizer=stdizer.Identity(),num_threads=num_threads)
            val = self._snpval_to_distval(snpdata.val,order,dtype)

            has_right_order = order="A" or (order=="C" and val.flags["C_CONTIGUOUS"]) or (order=="F" and val.flags["F_CONTIGUOUS"])
            assert has_right_order
            return val
        else: #Do in blocks
            t0 = time.time()
            if order=='A':
                order = 'F'
            val = np.zeros([self.iid_count,self.sid_count,3],dtype=dtype,order=order)#LATER use empty or fillnan???

            logging.info("reading {0} value data in blocks of {1} SNPs and finding distribution (for {2} individuals)".format(self.sid_count, self.block_size, self.iid_count))
            ct = 0
            ts = time.time()

            for start in range(0, self.sid_count, self.block_size):
                ct += self.block_size
                snpdata = self.snpreader[:,start:start+self.block_size].read(order=order,dtype=dtype,force_python_only=force_python_only,view_ok=True,num_threads=num_threads) # a view is always OK, because we'll allocate memory in the next step
                val[:,start:start+self.block_size] = self._snpval_to_distval(snpdata.val,order,dtype)
                if ct % self.block_size==0:
                    diff = time.time()-ts
                    if diff > 5:
                       logging.info("read %s SNPs in %.2f seconds" % (ct, diff))

            t1 = time.time()
            logging.info("%.2f seconds elapsed" % (t1-t0))

            return val


    def __getitem__(self, iid_indexer_and_snp_indexer):
        row_index_or_none, col_index_or_none = iid_indexer_and_snp_indexer
        return _Snp2Dist(self.snpreader[row_index_or_none,col_index_or_none],max_weight=self.max_weight,block_size=self.block_size)


    @property
    def sid(self):
        '''The :attr:`SnpReader.sid` property of the SNP data.
        '''
        return self.snpreader.sid

    @property
    def sid_count(self):
        '''The :attr:`SnpReader.sid_count` property of the SNP data.
        '''
        return self.snpreader.sid_count

    @property
    def pos(self):
        '''The :attr:`SnpReader.pos` property of the SNP data.
        '''
        return self.snpreader.pos


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod()
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import unittest
import logging
import os
import tempfile
import shutil
import time
from pysnptools.util.filecache import LocalCache, PeerToPeer, Hashdown


class TestFileCache(unittest.TestCase):

    def test_local_file(self):
        logging.info("test_local_file")

        temp_dir = self._temp_dir()
        def storage_closure():
            return LocalCache(temp_dir)
        self._write_and_read(storage_closure())
        self._distribute(storage_closure)
    
    def test_hashdown(self):
        logging.info("test_hashdown")

        from pysnptools.util.filecache.test import TestFileCache as self
        logging.basicConfig(level=logging.INFO)
        logging.info("test_hashdown")

        url='https://github.com/fastlmm/PySnpTools/raw/9de8e93a91b330b064b482c918a38104904b45c0/pysnptools'
        hashdown0 = Hashdown(url, allow_unknown_files=True)
        for file in ['examples/toydata.5chrom.bed','examples/toydata.bim','examples/toydata.fam','util/util.py']:
            hashdown0.file_exists(file)
        file_to_hash = hashdown0.file_to_hash
        hashdown = Hashdown(url, file_to_hash=hashdown0.file_to_hash, allow_unknown_files=False)

        #Clear the directory
        assert self._is_error(lambda : hashdown.rmtree()) # raise error because this is read-only
        #Rule: After you clear a directory, nothing is in it
        assert 4 == self._len(hashdown.walk()) #returns the files in the file_to_hash_file
        assert not hashdown.file_exists("test.txt")
        assert not hashdown.file_exists("main.txt/test.txt")
        assert not hashdown.file_exists(r"main.txt\test.txt")
        assert self._is_error(lambda : hashdown.file_exists("test.txt/")) #Can't query something that can't be a file name
        assert self._is_error(lambda : hashdown.file_exists("../test.txt")) #Can't leave the current directory
        if os.name == 'nt':
            assert self._is_error(lambda : hashdown.file_exists(r"c:\test.txt")) #Can't leave the current directory

        #Rule: '/' and '\' are both OK, but you can't use ':' or '..' to leave the current root.
        assert self._is_error(lambda : 0 == self._len(hashdown.walk("..")))
        assert 0 == self._len(hashdown.walk("..x"))
        assert 0 == self._len(hashdown.walk("test.txt")) #This is ok, because test.txt doesn't exist and therefore isn't a file
        assert 0 == self._len(hashdown.walk("a/b"))
        assert 0 == self._len(hashdown.walk("a\\b")) #Backslash or forward is fine
        assert self._is_error(lambda : len(hashdown.walk("/"))) #Can't start with '/'
        assert self._is_error(lambda : len(hashdown.walk(r"\\"))) #Can't start with '\'
        assert self._is_error(lambda : len(hashdown.walk(r"\\computer1\share\3"))) #Can't start with UNC


        #It should be there and be a file
        assert hashdown.file_exists('examples/toydata.bim')
        file_list = list(hashdown.walk())
        assert len(file_list)==4 and 'examples/toydata.bim' in file_list
        file_list2 = list(hashdown.walk("examples"))
        assert len(file_list2)==3 and 'examples/toydata.bim' in file_list2
        assert self._is_error(lambda : hashdown.join('examples/toydata.bim')) #Can't create a directory where a file exists
        assert self._is_error(lambda : list(hashdown.walk('examples/toydata.bim'))) #Can't have directory where a file exists

        #Read it
        assert hashdown.load('examples/toydata.bim').split('\n')[0] =="1\tnull_0\t0\t1\tL\tH"
        assert hashdown.file_exists('examples/toydata.bim')
        assert self._is_error(lambda : hashdown.load("examples"))  #This is an error because examples is actually a directory and they can't be opened for reading


        #Can query modified time of file.
        assert self._is_error(lambda : hashdown.getmtime("a/b/c.txt")), "Can't get mod time from file that doesn't exist"
        assert self._is_error(lambda : hashdown.getmtime("examples")), "Can't get mod time from directory"
        assert hashdown.getmtime('examples/toydata.bim') == 0.0, "expect all mod times to be 0"

        #try to write
        assert self._is_error(lambda : hashdown.save("main.txt","")), "Can't write. It's read only"
        #try to remove
        assert self._is_error(lambda : hashdown.remove('examples/toydata.bim')), "Can't remove. It's read only"
        assert self._is_error(lambda : hashdown.rmtree()), "Can't remove. It's read only"
        #what if files are not local?
        shutil.rmtree(hashdown.directory,ignore_errors=True)
        #It should be there and be a file
        assert hashdown.file_exists('examples/toydata.bim')
        file_list = list(hashdown.walk())
        assert len(file_list)==4 and 'examples/toydata.bim' in file_list
        file_list2 = list(hashdown.walk("examples"))
        assert len(file_list2)==3 and 'examples/toydata.bim' in file_list2
        assert hashdown.load('examples/toydata.bim').split('\n')[0] =="1\tnull_0\t0\t1\tL\tH"
        #What if put bad file locally?
        os.rename(hashdown.directory+'/examples/toydata.bim',hashdown.directory+'/examples/toydata.fam')
        assert hashdown.load('examples/toydata.fam').split('\n')[0] == 'per0 per0 0 0 0 0'
        #What if hash doesn't match web hash?
        hashdown.file_to_hash['examples/toydata.fam']='WRONG'
        assert self._is_error(lambda : hashdown.load('examples/toydata.fam')), "unexpected hash"
        #writing out file_to_hash and read_file_to_hash

        hashdown.save_hashdown('ignore/toydata.hashdown.json')
        hashdown2 = Hashdown.load_hashdown('ignore/toydata.hashdown.json')
        #It should be there and be a file
        assert hashdown2.file_exists('examples/toydata.bim')
        file_list = list(hashdown2.walk())
        assert len(file_list)==4 and 'examples/toydata.bim' in file_list
        file_list2 = list(hashdown2.walk("examples"))
        assert len(file_list2)==3 and 'examples/toydata.bim' in file_list2
        assert hashdown2.load('examples/toydata.bim').split('\n')[0] =="1\tnull_0\t0\t1\tL\tH"
        


    def test_peer_to_peer(self):
        from pysnptools.util.filecache import ip_address_pid
        logging.info("test_peer_to_peer")

        temp_dir = self._temp_dir()

        def storage_closure():
            def id_and_path_function():
                self.count += 1
                return self.count, temp_dir+'/{0}'.format(self.count)

            storage = PeerToPeer(common_directory=temp_dir+'/common',id_and_path_function=id_and_path_function)
            return storage

        self._write_and_read(storage_closure())
        self._distribute(storage_closure)
        

    @staticmethod
    def file_name(self,testcase_name):
            temp_fn = os.path.join(self.tempout_dir,testcase_name+".txt")
            if os.path.exists(temp_fn):
                os.remove(temp_fn)
            return temp_fn
    tempout_dir = "tempout/one_milp2p"

    @classmethod
    def setUpClass(self):
        self.temp_parent = os.path.join(tempfile.gettempdir(), format(hash(os.times())))
        self.count = 0

    def _temp_dir(self):
        self.count += 1
        return "{0}/{1}".format(self.temp_parent,self.count)

    @classmethod
    def tearDownClass(self):
        not os.path.exists(self.temp_parent) or shutil.rmtree(self.temp_parent)

    @staticmethod
    def _is_error(lambda0):
        try:
            lambda0()
        except Exception as e:
            logging.debug(str(e))
            return True
        return False

    @staticmethod
    def _len(sequence):
        len = 0
        for item in sequence:
            len += 1
        return len
        
    def _write_and_read(self,storage):
        test_storage = storage.join('test_snps') #!!!How to induce an error: create a 'test_snps' file at the top level then try to create an empty directory with the same name


        #Clear the directory
        test_storage.rmtree()
        #Rule: After you clear a directory, nothing is in it
        assert 0 == self._len(test_storage.walk())
        assert not test_storage.file_exists("test.txt")
        assert not test_storage.file_exists("main.txt/test.txt")
        assert not test_storage.file_exists(r"main.txt\test.txt")
        assert self._is_error(lambda : test_storage.file_exists("test.txt/")) #Can't query something that can't be a file name
        assert self._is_error(lambda : test_storage.file_exists("../test.txt")) #Can't leave the current directory
        if os.name == 'nt':
            assert self._is_error(lambda : test_storage.file_exists(r"c:\test.txt")) #Can't leave the current directory

        #Rule: '/' and '\' are both OK, but you can't use ':' or '..' to leave the current root.
        assert 0 == self._len(test_storage.walk())
        assert self._is_error(lambda : 0 == self._len(test_storage.walk("..")))
        assert 0 == self._len(test_storage.walk("..x"))
        assert 0 == self._len(test_storage.walk("test.txt")) #This is ok, because test.txt doesn't exist and therefore isn't a file
        assert 0 == self._len(test_storage.walk("a/b"))
        assert 0 == self._len(test_storage.walk("a\\b")) #Backslash or forward is fine
        assert self._is_error(lambda : len(test_storage.walk("/"))) #Can't start with '/'
        assert self._is_error(lambda : len(test_storage.walk(r"\\"))) #Can't start with '\'
        assert self._is_error(lambda : len(test_storage.walk(r"\\computer1\share\3"))) #Can't start with UNC

        #Clear the directory, again
        test_storage.rmtree()
        assert 0 == self._len(test_storage.walk())
        test_storage.rmtree("main.txt")
        assert 0 == self._len(test_storage.walk("main.txt"))
        assert 0 == self._len(test_storage.walk())


        #Write to it.
        assert self._is_error(lambda : test_storage.save("../test.txt"))
        test_storage.save("main.txt/test.txt","test\n")
        #Rule: It's an error to write to a file or directory that already exists
        assert self._is_error(lambda : test_storage.save("main.txt")) 
        assert self._is_error(lambda : test_storage.save("main.txt/test.txt")) 
        assert self._is_error(lambda : list(test_storage.walk("main.txt/test.txt"))), "Rule: It's an error to walk a file (but recall that it's OK to walk a folder that doesn't exist)"

        #It should be there and be a file
        assert test_storage.file_exists("main.txt/test.txt")
        file_list = list(test_storage.walk())
        assert len(file_list)==1 and file_list[0] == "main.txt/test.txt"
        file_list2 = list(test_storage.walk("main.txt"))
        assert len(file_list2)==1 and file_list2[0] == "main.txt/test.txt"
        assert self._is_error(lambda : test_storage.join("main.txt/test.txt")) #Can't create a directory where a file exists
        assert self._is_error(lambda : list(test_storage.walk("main.txt/test.txt"))) #Can't create a directory where a file exists
        assert self._is_error(lambda : test_storage.rmtree("main.txt/test.txt")) #Can't delete a directory where a file exists

        #Read it
        assert test_storage.load("main.txt/test.txt")=="test\n"
        assert test_storage.file_exists("main.txt/test.txt")
        assert self._is_error(lambda : test_storage.load("main.txt"))  #This is an error because main.txt is actually a directory and they can't be opened for reading

        #Remove it
        test_storage.remove("main.txt/test.txt")
        assert self._is_error(lambda : test_storage.remove("main.txt/test.txt")) #Can't delete a file that doesn't exist
        assert not test_storage.file_exists("main.txt/test.txt")
        assert 0 == self._len(test_storage.walk())
        assert 0 == self._len(test_storage.walk("main.txt"))
        assert 0 == self._len(test_storage.walk("main.txt/test.txt")) #Now allowed.

        #  writing zero length files is OK
        #  File share has a special file called "main.txt". Can we mess things up by using 'main.txt' as a directory name, too.
        #It's OK to write to a file in a directory that used to exist, but now has no files.
        test_storage.save("main.txt","")
        assert test_storage.file_exists("main.txt")
        file_list = list(test_storage.walk())
        assert len(file_list)==1 and file_list[0] == "main.txt"
        assert test_storage.load("main.txt") == ""
        assert test_storage.file_exists("main.txt")

        #Can query modified time of file. It will be later, later.
        assert self._is_error(lambda : test_storage.getmtime("a/b/c.txt")), "Can't get mod time from file that doesn't exist"
        test_storage.save("a/b/c.txt","")
        m1 = test_storage.getmtime("a/b/c.txt")
        assert self._is_error(lambda : test_storage.getmtime("a/b")), "Can't get mod time from directory"
        assert test_storage.getmtime("a/b/c.txt") == m1, "expect mod time to stay the same"
        assert test_storage.load("a/b/c.txt") == ""
        assert test_storage.getmtime("a/b/c.txt") == m1, "reading a file doesn't change its mod time"
        test_storage.remove("a/b/c.txt")
        assert self._is_error(lambda : test_storage.getmtime("a/b/c.txt")), "Can't get mod time from file that doesn't exist"
        time.sleep(1) #Sleep one second
        test_storage.save("a/b/c.txt","")
        test_storage.save("a/b/d.txt","")
        assert test_storage.getmtime("a/b/c.txt") > m1, "A file created later (after a pause) will have a later mod time"
        assert test_storage.getmtime("a/b/d.txt") > m1, "A file created later (after a pause) will have a later mod time"
        assert test_storage.getmtime("a/b/d.txt") >= test_storage.getmtime("a/b/c.txt"), "A file created later (with no pause) will have a later or equal mod time"

    # Look for code that isn't covered and make test cases for it
    #Distributed testing
    # What happens if you don't close an open_write? Undefined
    
    def _distribute(self,storage_lambda):
        storage1 = storage_lambda()
        storage2 = storage_lambda()
        #clear everything on #1
        storage1.rmtree()
        #1 and #2 agree nothing is there
        assert 0 == self._len(storage1.walk())
        assert 0 == self._len(storage2.walk())
        #write on #1
        storage1.save("a/b/c.txt","Hello")
        #read on #2
        assert storage2.load("a/b/c.txt")=="Hello"
        #read on #2, again and manually see that it caches
        assert storage2.load("a/b/c.txt")=="Hello"
        #remove on #1
        storage1.remove("a/b/c.txt")
        #assert not exists on #2
        assert not storage2.file_exists("a/b/c.txt")
        #write on #1
        storage1.save("a/b/c.txt","Hello")
        #read on #2
        assert storage2.load("a/b/c.txt")=="Hello"
        #remove on #1
        storage1.remove("a/b/c.txt")
        #write something different on #1
        storage1.save("a/b/c.txt","There")
        #read on #2 and see that it is different.
        assert storage2.load("a/b/c.txt")=="There"

    def test_util_filecache_testmod(self):
        import doctest
        import pysnptools.util.filecache
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        for mod in [
                    pysnptools.util.filecache,
                    pysnptools.util.filecache.filecache,
                    pysnptools.util.filecache.localcache,
                    pysnptools.util.filecache.peertopeer,
                    pysnptools.util.filecache.hashdown,
                    ]:
            result = doctest.testmod(mod,optionflags=doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE)
            assert result.failed == 0, "failed doc test: " + __file__
        os.chdir(old_dir)


def getTestSuite():
    """
    set up composite test suite
    """
    
    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestFileCache))
    return test_suite

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    suites = getTestSuite()

    r = unittest.TextTestRunner(failfast=True)#!!! should be false
    ret = r.run(suites)
    assert ret.wasSuccessful()

    logging.info("done")



import logging
import os
import numpy as np
from pysnptools.pstreader import PstReader


#!!! would be better to make a Transpose class that could term _mergerows into mergecols? Be sure special Bed code is still there.
class _MergeCols(PstReader):
    def __init__(self,reader_list,cache_file=None,skip_check=False):
        super(_MergeCols, self).__init__()
        assert len(reader_list) > 0, "Expect at least one reader"
        self.skip_check = skip_check

        self.reader_list = reader_list
        self._repr_string = "_MergeCols({0})".format(",".join([str(s) for s in reader_list]))

        if cache_file is not None:
            #!!!add warning if cache_file doesn't end with .npz
            if not os.path.exists(cache_file):
                self._run_once()
                self._savez(cache_file)
            else:
                self._load(cache_file)

        def _savez(self, cache_file):
            np.savez(cache_file,
                     _row=self._row, _row_property=self._row_property,
                     _col=self._col, _col_property=self._col_property,
                     col_count_list=self.col_count_list)

        def _load(self,cache_file):
                with np.load(cache_file,allow_pickle=True) as data:
                    self._col = data['_col']
                    self._col_property = data['_col_property']
                    self.col_count_list = data['col_count_list']
                    assert ('_row' in data) == ('_row_property' in data)
                    self._row = data['_row']
                    self._row_property = data['_row_property']

    def _run_once(self):
        if hasattr(self,'_has_run_once'):
            return
        from pysnptools.snpreader import Bed
        self._has_run_once = True
        #Check that all rows are distinct and that all cols and pos are the same and in the same order

        if self.skip_check and all(isinstance(reader,Bed) for reader in self.reader_list): #Special code if all Bed readers
            l = [SnpReader._read_map_or_bim(reader.filename,remove_suffix="bed", add_suffix="bim") for reader in self.reader_list]
            self.col_count_list = np.array([len(ll[0]) for ll in l])
            self._row = self.reader_list[0].row
            self._row_property = self.reader_list[0].row_property
            self._col = np.concatenate([ll[0] for ll in l])
            self._col_property = np.concatenate([ll[1] for ll in l])
        else:
            col_list = []
            col_property_list = []
            col_set = set()
            self.col_count_list = []
            self._row = self.reader_list[0].row
            self._row_property = self.reader_list[0].row_property
            for reader_index,reader in enumerate(self.reader_list):
                if reader_index % 10 == 0: logging.info("_MergeCols looking at reader #{0}: {1}".format(reader_index,reader))
                if not self.skip_check:
                    assert np.array_equal(self._row,reader.row), "Expect rows to be the same across all files"
                    np.testing.assert_equal(self._row_property,reader.row_property) #"Expect column_property to be the same across all files"
                    size_before = len(col_set)
                    col_set.update((tuple(item) for item in reader.col))
                    assert len(col_set) == size_before + reader.col_count, "Expect cols to be distinct in all files"
                col_list.append(reader.col)
                col_property_list.append(reader.col_property)
                self.col_count_list.append(reader.col_count)
            self._col = np.concatenate(col_list)
            self._col_property = np.concatenate(col_property_list)
            self.col_count_list = np.array(self.col_count_list)

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    def __repr__(self): 
        #Don't need _run_once because based only on initial info
        return self._repr_string

    def copyinputs(self, copier):
        self._run_once()
        for reader in self.reader_list:
            copier.input(reader)

    def _create_reader_and_col_index_list(self,col_index):
        result = []
        start = 0
        for reader_index in range(len(self.reader_list)):
            stop = start + self.col_count_list[reader_index] #!!! shouldn't this be col_count (and check _mergerows, too)
            is_here = (col_index >= start) * (col_index < stop)
            if any(is_here):
                col_index_rel = col_index[is_here]-start
                result.append((reader_index,is_here,col_index_rel))
            start = stop
        return result

    def _read(self, row_index_or_none, col_index_or_none, order, dtype, force_python_only, view_ok, num_threads):
        #!!!tests to do: no row's
        #!!!tests to do: no col's
        #!!!test to do: from file 1, file2, and then file1 again
        dtype = np.dtype(dtype)
        col_index = col_index_or_none if col_index_or_none is not None else np.arange(self.col_count) #!!!might want to special case reading all
        row_index_or_none_count = self.row_count if row_index_or_none is None else len(row_index_or_none)
        col_index_or_none_count = self.col_count if col_index_or_none is None else len(col_index_or_none)

        #Create a list of (reader,col_index)
        reader_and_col_index_list = self._create_reader_and_col_index_list(col_index)
        if len(reader_and_col_index_list) == 0:
            return self.reader_list[0]._read(row_index_or_none,col_index,order,dtype,force_python_only, view_ok, num_threads)
        elif len(reader_and_col_index_list) == 1:
            reader_index,col_index_in,col_index_rel = reader_and_col_index_list[0]
            reader = self.reader_list[reader_index]
            return reader._read(row_index_or_none,col_index_rel,order,dtype,force_python_only, view_ok, num_threads)
        else:
            logging.info("Starting read from {0} subreaders".format(len(reader_and_col_index_list)))
            if order == 'A' or order is None:#LATER does every _read( need code like this?
                order = 'F'
            val = None

            for reader_index,is_here,col_index_rel in reader_and_col_index_list:
                reader = self.reader_list[reader_index]
                if reader_index % 1 == 0: logging.info("Reading from #{0}: {1}".format(reader_index,reader))
                piece = reader._read(row_index_or_none,col_index_rel,order,dtype,force_python_only, view_ok=True, num_threads=num_threads)
                if val is None:
                    if len(piece.shape) == 2:
                        val = np.empty([row_index_or_none_count, col_index_or_none_count], dtype=dtype, order=order)
                    else:
                        val = np.empty([row_index_or_none_count, col_index_or_none_count, self.val_shape], dtype=dtype, order=order)
                    val.fill(np.nan) #LATER Keep this?

                if len(val.shape)== 2:
                    val[:,is_here] = piece 
                else:
                    val[:,is_here,:] = piece

            logging.info("Ended read from {0} subreaders".format(len(reader_and_col_index_list)))
            assert val is not None
            return val




import numpy as np
import scipy as sp
import logging
import doctest
import shutil
import unittest
import os.path
import time
import sys


from pysnptools.distreader.distmemmap import TestDistMemMap
try:
    from bgen_reader import open_bgen
    from pysnptools.distreader.bgen import TestBgen
    _TEST_BGEN_AVAILABLE = True
except:
    logging.warning("BgenReader will not be available")
    _TEST_BGEN_AVAILABLE = False
from pysnptools.distreader.distgen import TestDistGen
from pysnptools.distreader import DistNpz, DistHdf5, DistMemMap, DistData, _DistMergeSIDs
from pysnptools.util import create_directory_if_necessary
from pysnptools.snpreader import SnpNpz, Bed
from pysnptools.kernelreader.test import _fortesting_JustCheckExists
from pysnptools.pstreader import PstData
from pysnptools.snpreader import SnpData

class TestDistReaders(unittest.TestCase):     


    @classmethod
    def setUpClass(self):
        self.currentFolder = os.path.dirname(os.path.realpath(__file__))
        #TODO: get data set with NANs!
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        self.distdata = distreader.read(order='F',force_python_only=True)
        self.dist_values = self.distdata.val

    def test_3d(self):
        from pysnptools.distreader import DistData
        np.random.seed(0)
        row_count = 4
        col_count = 5
        val_shape = 3
        val = np.random.random((row_count,col_count,val_shape))
        val /= val.sum(axis=2,keepdims=True)  #make probabilities sum to 1
        distdata = DistData(val=val,iid=[['iid{0}'.format(i)]*2 for i in range(row_count)],sid=['sid{0}'.format(s) for s in range(col_count)]
                            )

    def test_scalar_index(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        arr=np.int64(1)
        distreader[arr,arr]

    def test_c_reader_hdf5(self):
        distreader = DistHdf5(self.currentFolder + "/../examples/toydata.snpmajor.dist.hdf5")
        self.c_reader(distreader)

    def test_hdf5_case3(self):
        distreader1 = DistHdf5(self.currentFolder + "/../examples/toydata.snpmajor.dist.hdf5")[::2,:]
        distreader2 = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")[::2,:]
        self.assertTrue(np.allclose(distreader1.read().val, distreader2.read().val, rtol=1e-05, atol=1e-05))


    def test_c_reader_npz(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata10.dist.npz")
        distdata = distreader.read(order='F',force_python_only=False)
        snp_c = distdata.val
        
        self.assertEqual(np.float64, snp_c.dtype)
        self.assertTrue(np.allclose(self.dist_values[:,:10], snp_c, rtol=1e-05, atol=1e-05))

        distreader1 = DistNpz(self.currentFolder + "/../examples/toydata10.dist.npz")
        distreader2 = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")[:,:10]
        self.assertTrue(np.allclose(distreader1.read().val, distreader2.read().val, rtol=1e-05, atol=1e-05))


        distdata.val[1,2] = np.NaN # Inject a missing value to test writing and reading missing values
        output = "tempdir/distreader/toydata10.dist.npz"
        create_directory_if_necessary(output)
        DistNpz.write(output,distdata)
        snpdata2 = DistNpz(output).read()
        np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)

        snpdata3 = distdata[:,0:0].read() #create distdata with no sids
        output = "tempdir/distreader/toydata0.dist.npz"
        DistNpz.write(output,snpdata3)
        snpdata4 = DistNpz(output).read()
        assert snpdata3 == snpdata4

               
    def c_reader(self,distreader):
        """
        make sure c-reader yields same result
        """
        distdata = distreader.read(order='F',force_python_only=False)
        snp_c = distdata.val
        
        self.assertEqual(np.float64, snp_c.dtype)
        self.assertTrue(np.allclose(self.dist_values, snp_c, rtol=1e-05, atol=1e-05))
        return distdata

    def test_write_distnpz_f64cpp_0(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        iid_index = 0
        logging.info("iid={0}".format(iid_index))
        #if distreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    distreader = distreader[0:-1,:]
        #assert distreader.iid_count % 4 != 0
        distdata = distreader[0:iid_index,:].read(order='F',dtype=np.float64)
        if distdata.iid_count > 0:
            distdata.val[-1,0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}.dist.npz".format(iid_index)
        create_directory_if_necessary(output)
        DistNpz.write(output, distdata )
        snpdata2 = DistNpz(output).read()
        np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)

    def test_write_distnpz_f64cpp_1(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        iid_index = 1
        logging.info("iid={0}".format(iid_index))
        #if distreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    distreader = distreader[0:-1,:]
        #assert distreader.iid_count % 4 != 0
        distdata = distreader[0:iid_index,:].read(order='F',dtype=np.float64)
        if distdata.iid_count > 0:
            distdata.val[-1,0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}.dist.npz".format(iid_index)
        create_directory_if_necessary(output)
        DistNpz.write(output, distdata )
        snpdata2 = DistNpz(output).read()
        np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)

    def test_write_distnpz_f64cpp_5(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")

        _fortesting_JustCheckExists().input(distreader)

        iid_index = 5
        logging.info("iid={0}".format(iid_index))
        #if distreader.iid_count % 4 == 0: # divisible by 4 isn't a good test
        #    distreader = distreader[0:-1,:]
        #assert distreader.iid_count % 4 != 0
        distdata = distreader[0:iid_index,:].read(order='F',dtype=np.float64)
        if distdata.iid_count > 0:
            distdata.val[-1,0] = float("NAN")
        output = "tempdir/toydata.F64cpp.{0}.dist.npz".format(iid_index)
        create_directory_if_necessary(output)
        DistNpz.write(output, distdata ) #,force_python_only=True)
        snpdata2 = DistNpz(output).read()
        np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)



    def test_write_x_x_cpp(self):
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        for order in ['C','F']:
            for dtype in [np.float32,np.float64]:
                distdata = distreader.read(order=order,dtype=dtype)
                distdata.val[-1,0] = float("NAN")
                output = "tempdir/toydata.{0}{1}.cpp.dist.npz".format(order,"32" if dtype==np.float32 else "64")
                create_directory_if_necessary(output)
                DistNpz.write(output, distdata)
                snpdata2 = DistNpz(output).read()
                np.testing.assert_array_almost_equal(distdata.val, snpdata2.val, decimal=10)

    def test_subset_view(self):
        distreader2 = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")[:,:]
        result = distreader2.read(view_ok=True)
        self.assertFalse(distreader2 is result)
        result2 = result[:,:].read()
        self.assertFalse(np.may_share_memory(result2.val,result.val))
        result3 = result[:,:].read(view_ok=True)
        self.assertTrue(np.may_share_memory(result3.val,result.val))
        result4 = result3.read()
        self.assertFalse(np.may_share_memory(result4.val,result3.val))
        result5 = result4.read(view_ok=True)
        self.assertTrue(np.may_share_memory(result4.val,result5.val))




    def test_writes(self):
        from pysnptools.distreader import DistData, DistHdf5, DistNpz, DistMemMap
        from pysnptools.kernelreader.test import _fortesting_JustCheckExists

        the_class_and_suffix_list = [(DistNpz,"npz",None,None),
                                     # Bgen used to be here
                                     (DistHdf5,"hdf5",None,None),
                                     (DistMemMap,"memmap",None,None)]
        # Skip because write requires qctool, which is not installed by default
        # if _TEST_BGEN_AVAILABLE:
        #     from pysnptools.distreader import Bgen
        #     the_class_and_suffix_list.append((Bgen,"bgen",None,lambda filename,distdata: Bgen.write(filename,distdata,bits=32)))

        cant_do_col_prop_none_set = {'bgen'}
        cant_do_col_len_0_set = {'bgen'}
        cant_do_row_count_zero_set = {'bgen'}
        can_swap_0_2_set = {}
        can_change_col_names_set = {}
        ignore_fam_id_set = {}
        ignore_pos1_set = {'bgen'}
        ignore_pos_set = {}
        erase_any_write_dir = {}

        
        #===================================
        #    Starting main function
        #===================================
        logging.info("starting 'test_writes'")
        np.random.seed(0)
        output_template = "tempdir/distreader/writes.{0}.{1}"
        create_directory_if_necessary(output_template.format(0,"npz"))
        i = 0
        for row_count in [0,5,2,1]:
            for col_count in [4,2,1,0]:
                val=np.random.random(size=[row_count,col_count,3])
                val /= val.sum(axis=2,keepdims=True)  #make probabilities sum to 1

                val[val==3]=np.NaN
                row = [('0','0'),('1','1'),('2','2'),('3','3'),('4','4')][:row_count]
                col = ['s0','s1','s2','s3','s4'][:col_count]
                for is_none in [True,False]:
                    row_prop = None
                    col_prop = None if is_none else [(x,x,x) for x in range(5)][:col_count]
                    distdata = DistData(iid=row,sid=col,val=val,pos=col_prop,name=str(i))
                    for the_class,suffix,constructor,writer in the_class_and_suffix_list:
                        constructor = constructor or (lambda filename: the_class(filename))
                        writer = writer or (lambda filename,distdata: the_class.write(filename,distdata))
                        if col_count == 0 and suffix in cant_do_col_len_0_set:
                            continue
                        if col_prop is None and suffix in cant_do_col_prop_none_set:
                            continue
                        if row_count==0 and suffix in cant_do_row_count_zero_set:
                            continue
                        filename = output_template.format(i,suffix)
                        logging.info(filename)
                        i += 1
                        if suffix in erase_any_write_dir and os.path.exists(filename):
                            shutil.rmtree(filename)
                        ret = writer(filename,distdata)
                        assert ret is not None
                        for subsetter in [None, np.s_[::2,::3]]:
                            reader = constructor(filename)
                            _fortesting_JustCheckExists().input(reader)
                            subreader = reader if subsetter is None else reader[subsetter[0],subsetter[1]]
                            readdata = subreader.read(order='C')
                            expected = distdata if subsetter is None else distdata[subsetter[0],subsetter[1]].read()
                            if not suffix in can_swap_0_2_set:
                                assert np.allclose(readdata.val,expected.val,equal_nan=True)
                            else:
                                for col_index in range(readdata.col_count):
                                    assert (np.allclose(readdata.val[:,col_index],expected.val[:,col_index],equal_nan=True) or
                                            np.allclose(readdata.val[:,col_index]*-1+2,expected.val[:,col_index],equal_nan=True))
                            if not suffix in ignore_fam_id_set:
                                assert np.array_equal(readdata.row,expected.row)
                            else:
                                assert np.array_equal(readdata.row[:,1],expected.row[:,1])
                            if not suffix in can_change_col_names_set:
                                assert np.array_equal(readdata.col,expected.col)
                            else:
                                assert readdata.col_count==expected.col_count
                            assert np.array_equal(readdata.row_property,expected.row_property) or (readdata.row_property.shape[1]==0 and expected.row_property.shape[1]==0)

                            if suffix in ignore_pos1_set:
                                assert np.allclose(readdata.col_property[:,[0,2]],expected.col_property[:,[0,2]],equal_nan=True) or (readdata.col_property.shape[1]==0 and expected.col_property.shape[1]==0)
                            elif not suffix in ignore_pos_set:
                                assert np.allclose(readdata.col_property,expected.col_property,equal_nan=True) or (readdata.col_property.shape[1]==0 and expected.col_property.shape[1]==0)
                            else:
                                assert len(readdata.col_property)==len(expected.col_property)
                        try:
                            os.remove(filename)
                        except:
                            pass
        logging.info("done with 'test_writes'")

    def test_block_size_DistData(self):
        np.random.seed(0)
        sid_count = 20
        val=np.array(np.random.random(size=[3,sid_count,3]),dtype=np.float64,order='F')
        val /= val.sum(axis=2,keepdims=True)  #make probabilities sum to 1
        distreader = DistData(iid=[["0","0"],["1","1"],["2","2"]],sid=[str(i) for i in range(sid_count)],val=val)
        snpdata0 = distreader.as_snp(max_weight=100,block_size=1).read()
        snpdata1 = distreader.as_snp(max_weight=100,block_size=None).read()
        np.testing.assert_array_almost_equal(snpdata0.val,snpdata1.val, decimal=10)

    def test_block_size_Snp2Dist(self):
        from pysnptools.snpreader import SnpData
        from pysnptools.distreader._snp2dist import _Snp2Dist
        np.random.seed(0)
        sid_count = 20
        val=np.array(np.random.randint(0,3,size=[3,sid_count]),dtype=np.float64,order='F')
        snpreader = SnpData(iid=[["0","0"],["1","1"],["2","2"]],sid=[str(i) for i in range(sid_count)],val=val)
        distdata0 = snpreader.as_dist(max_weight=2,block_size=1).read()
        distdata1 = snpreader.as_dist(max_weight=2,block_size=None).read()
        np.testing.assert_array_almost_equal(distdata0.val,distdata1.val, decimal=10)

    def test_intersection_Dist2Snp(self):
        from pysnptools.snpreader._dist2snp import _Dist2Snp
        from pysnptools.snpreader import Pheno
        from pysnptools.distreader._subset import _DistSubset
        from pysnptools.snpreader._subset import _SnpSubset
        from pysnptools.util import intersect_apply

        dist_all = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        k = dist_all.as_snp(max_weight=25)

        pheno = Pheno(self.currentFolder + "/../examples/toydata.phe")
        pheno = pheno[1:,:] # To test intersection we remove a iid from pheno

        k1,pheno = intersect_apply([k,pheno]) 
        assert isinstance(k1.distreader,_DistSubset) and not isinstance(k1,_SnpSubset)

        #What happens with fancy selection?
        k2 = k[::2,:]
        assert isinstance(k2,_Dist2Snp)

        logging.info("Done with test_intersection")

    def test_intersection_Snp2Dist(self):
        from pysnptools.distreader._snp2dist import _Snp2Dist
        from pysnptools.snpreader import Pheno, Bed
        from pysnptools.distreader._subset import _DistSubset
        from pysnptools.snpreader._subset import _SnpSubset
        from pysnptools.util import intersect_apply

        snp_all = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=True)
        k = snp_all.as_dist(max_weight=2)

        pheno = Pheno(self.currentFolder + "/../examples/toydata.phe")
        pheno = pheno[1:,:] # To test intersection we remove a iid from pheno

        k1,pheno = intersect_apply([k,pheno]) 
        assert isinstance(k1.snpreader,_SnpSubset) and not isinstance(k1,_DistSubset)

        #What happens with fancy selection?
        k2 = k[::2,:]
        assert isinstance(k2,_Snp2Dist)

        logging.info("Done with test_intersection")

    def test_roundtrip(self):
        max_weight = 2
        snpreader1 = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=True)
        snpdata1 = snpreader1.read()
        distreader1 = snpreader1.as_dist(max_weight)
        snpreader2 = distreader1.as_snp(max_weight)
        assert snpdata1.allclose(snpreader2.read(),equal_nan=True)
        snpdata1.val[0,0] = np.nan
        assert snpdata1.allclose(snpdata1.as_dist(max_weight).as_snp(max_weight).read(),equal_nan=True)

    def test_dist_snp2(self):
        logging.info("in test_dist_snp2")
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        dist2snp = distreader.as_snp(max_weight=33)
        s = str(dist2snp)
        _fortesting_JustCheckExists().input(dist2snp)

    def test_snp_dist2(self):
        logging.info("in test_snp_dist2")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        snp2dist = snpreader.as_dist(max_weight=2)
        s = str(snp2dist)
        _fortesting_JustCheckExists().input(snp2dist)

    def test_subset_Dist2Snp(self):
        logging.info("in test_subset")
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        dist2snp = distreader.as_snp(max_weight=10)
        dssub = dist2snp[::2,::2]
        snpdata1 = dssub.read()
        expected = distreader.as_snp(max_weight=10)[::2,::2].read()
        np.testing.assert_array_almost_equal(snpdata1.val, expected.val, decimal=10)

        logging.info("done with test")

    def test_subset_Snp2Dist(self): #!!!move these to another test class
        logging.info("in test_subset")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        snp2dist = snpreader.as_dist(max_weight=2)
        sub = snp2dist[::2,::2]
        distdata1 = sub.read()
        expected = snpreader.as_dist(max_weight=2)[::2,::2].read()
        np.testing.assert_array_almost_equal(distdata1.val, expected.val, decimal=10)

        logging.info("done with test")

    def test_respect_read_inputs(self):
        from pysnptools.distreader import DistGen,DistHdf5,DistMemMap,DistNpz
        if _TEST_BGEN_AVAILABLE:
            from pysnptools.distreader import Bgen
        from pysnptools.snpreader import Bed

        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        distreader_list = [Bed('../examples/toydata.5chrom.bed',count_A1=True).as_dist(block_size=2000),
                           Bed('../examples/toydata.5chrom.bed',count_A1=True).as_dist(),
                           DistGen(seed=0,iid_count=500,sid_count=50),
                           DistGen(seed=0,iid_count=500,sid_count=50)[::2,::2],
                           DistHdf5('../examples/toydata.snpmajor.dist.hdf5'),
                           DistMemMap('../examples/tiny.dist.memmap'),
                           DistNpz('../examples/toydata10.dist.npz')
                          ]
        if _TEST_BGEN_AVAILABLE:
            distreader_list += [_DistMergeSIDs([Bgen('../examples/example.bgen')[:,:5].read(),Bgen('../examples/example.bgen')[:,5:].read()]),
                                Bgen('../examples/example.bgen').read(),
                                Bgen('../examples/bits1.bgen')]

        for distreader in distreader_list:
            logging.info(str(distreader))
            for order in ['F','C','A']:
                for dtype in [np.float32,np.float64]:
                    for force_python_only in [True,False]:
                        for view_ok in [True,False]:
                            val = distreader.read(order=order,dtype=dtype,force_python_only=force_python_only,view_ok=view_ok).val
                            has_right_order = order=="A" or (order=="C" and val.flags["C_CONTIGUOUS"]) or (order=="F" and val.flags["F_CONTIGUOUS"])
                            if hasattr(distreader,'val') and not view_ok:
                                assert distreader.val is not val
                            if (hasattr(distreader,'val') and view_ok and distreader.val is not val and
                                (order == 'A' or (order == 'F' and distreader.val.flags['F_CONTIGUOUS']) or (order == 'C' and distreader.val.flags['C_CONTIGUOUS'])) and
                                (dtype is None or  distreader.val.dtype == dtype)):
                                logging.info("{0} could have read a view, but didn't".format(distreader))
                            assert val.dtype == dtype and has_right_order

        os.chdir(previous_wd)


    def test_respect_inputs_DistData(self):
        np.random.seed(0)
        for dtype_start,decimal_start in [(np.float32,5),(np.float64,10)]:
            for order_start in ['F','C','A']:
                for sid_count in [20,2]:
                    val=np.array(np.random.random(size=[3,sid_count,3]),dtype=dtype_start,order=order_start)
                    val /= val.sum(axis=2,keepdims=True)  #make probabilities sum to 1
                    distdataX = DistData(iid=[["0","0"],["1","1"],["2","2"]],sid=[str(i) for i in range(sid_count)],val=val)
                    for max_weight in [1.0,2.0]:
                        weights = np.array([0,.5,1])*max_weight
                        for distreader0 in [distdataX,distdataX[:,1:]]:
                            distreader1 = distreader0[1:,:]
                            refdata0 = distreader0.read()
                            refval0 = (refdata0.val * weights).sum(axis=-1)
                            for dtype_goal,decimal_goal in [(np.float32,5),(np.float64,10)]:
                                for order_goal in ['F','C','A']:
                                    k = distreader0.as_snp(max_weight=max_weight,block_size=1).read(order=order_goal,dtype=dtype_goal)
                                    DistData._array_properties_are_ok(k.val,order_goal,dtype_goal)
                                    np.testing.assert_array_almost_equal(refval0,k.val, decimal=min(decimal_start,decimal_goal))


    def test_npz(self):
        logging.info("in test_npz")
        distreader = DistNpz(self.currentFolder + "/../examples/toydata.dist.npz")
        snpdata1 = distreader.as_snp(max_weight=1.0).read()
        s = str(snpdata1)
        output = "tempdir/distreader/toydata.snp.npz"
        create_directory_if_necessary(output)
        SnpNpz.write(output,snpdata1)
        snpreader2 = SnpNpz(output)
        snpdata2 = snpreader2.read()
        np.testing.assert_array_almost_equal(snpdata1.val, snpdata2.val, decimal=10)
        logging.info("done with test")

class TestDistNaNCNC(unittest.TestCase):
    def __init__(self, iid_index_list, snp_index_list, distreader, dtype, order, force_python_only, reference_snps, reference_dtype):
        self.iid_index_list = iid_index_list
        self.snp_index_list = snp_index_list
        self.distreader = distreader
        self.dtype = np.dtype(dtype)
        self.order = order
        self.force_python_only = force_python_only
        self.reference_snps = reference_snps
        self.reference_dtype = reference_dtype

    _testMethodName = "runTest"
    _testMethodDoc = None

    @staticmethod
    def factory_iterator():

        snp_reader_factory_distnpz = lambda : DistNpz("../examples/toydata.dist.npz")
        snp_reader_factory_snpmajor_hdf5 = lambda : DistHdf5("../examples/toydata.snpmajor.dist.hdf5")
        snp_reader_factory_iidmajor_hdf5 = lambda : DistHdf5("../examples/toydata.iidmajor.dist.hdf5")

        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        distreader0 = snp_reader_factory_distnpz()
        S_original = distreader0.sid_count
        N_original = distreader0.iid_count

        snps_to_read_count = min(S_original, 100)

        for iid_index_list in [range(N_original), range(N_original//2), range(N_original - 1,0,-2)]:
            for snp_index_list in [range(snps_to_read_count), range(snps_to_read_count//2), range(snps_to_read_count - 1,0,-2)]:
                reference_snps, reference_dtype = TestDistNaNCNC(iid_index_list, snp_index_list, snp_reader_factory_distnpz(), np.float64, "C", "False", None, None).read_and_standardize()
                for distreader_factory in [snp_reader_factory_distnpz, 
                                            snp_reader_factory_snpmajor_hdf5, snp_reader_factory_iidmajor_hdf5
                                            ]:
                    for dtype in [np.float64,np.float32]:
                        for order in ["C", "F"]:
                            for force_python_only in [False, True]:
                                distreader = distreader_factory()
                                test_case = TestDistNaNCNC(iid_index_list, snp_index_list, distreader, dtype, order, force_python_only, reference_snps, reference_dtype)
                                yield test_case
        os.chdir(previous_wd)

    def __str__(self):
        iid_index_list = self.iid_index_list
        snp_index_list = self.snp_index_list
        distreader = self.distreader
        dtype = self.dtype
        order = self.order
        force_python_only = self.force_python_only
        return "{0}(iid_index_list=[{1}], snp_index_list=[{2}], distreader={3}, dtype={4}, order='{5}', force_python_only=={6})".format(
            self.__class__.__name__,
            ",".join([str(i) for i in iid_index_list]) if len(iid_index_list) < 10 else ",".join([str(i) for i in iid_index_list[0:10]])+",...",
            ",".join([str(i) for i in snp_index_list]) if len(snp_index_list) < 10 else ",".join([str(i) for i in snp_index_list[0:10]])+",...",
            distreader, dtype, order, force_python_only)

    def read_and_standardize(self):
        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        iid_index_list = self.iid_index_list
        snp_index_list = self.snp_index_list
        distreader = self.distreader
        dtype = self.dtype
        order = self.order
        force_python_only = self.force_python_only
        
        snps = distreader[iid_index_list,snp_index_list].read(order=order, dtype=dtype, force_python_only=force_python_only).val
        snps[0,0] = [np.nan]*3 # add a NaN
        snps[:,1] = [.1,.2,.7] # make a SNC
        os.chdir(previous_wd)
        return snps, dtype

    def doCleanups(self):
        pass
        #return super(NaNCNCTestCases, self).doCleanups()

    def runTest(self, result = None):
        snps, dtype = self.read_and_standardize()
        assert not np.array_equal(snps[0,0],snps[0,0]) #without SnpReader's standardization NaN's stay NaN's
        assert np.allclose(snps[0,1],[.1,.2,.7])
        if self.reference_snps is not None:
            self.assertTrue(np.allclose(self.reference_snps, snps, rtol=1e-04 if dtype == np.float32 or self.reference_dtype == np.float32 else 1e-12,equal_nan=True))


# We do it this way instead of using doctest.DocTestSuite because doctest.DocTestSuite requires modules to be pickled, which python doesn't allow.
# We need tests to be pickleable so that they can be run on a cluster.
class TestDistReaderDocStrings(unittest.TestCase):
    def test_distreader(self):
        import pysnptools.distreader.distreader
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.distreader.distreader,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_distdata(self):
        import pysnptools.distreader.distdata
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.distreader.distdata,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


    def test_disthdf5(self):
        import pysnptools.distreader.disthdf5
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.distreader.disthdf5,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_distnpz(self):
        import pysnptools.distreader.distnpz
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.distreader.distnpz,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])

    if _TEST_BGEN_AVAILABLE:
        test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestBgen))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistReaderDocStrings))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistGen))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistMemMap))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistReaders))
    test_suite.addTests(TestDistNaNCNC.factory_iterator())
    

    return test_suite

if __name__ == '__main__':
    logging.basicConfig(level=logging.WARN)

    if False:
        from pysnptools.snpreader import Bed
        from pysnptools.distreader import DistData, DistNpz
        # Create toydata.dist.npz
        currentFolder = os.path.dirname(os.path.realpath(__file__))
        if True:
            snpreader = Bed(currentFolder + "/../examples/toydata.5chrom.bed",count_A1=True)[:25,:]
            np.random.seed(392)
            val = np.random.random((snpreader.iid_count,snpreader.sid_count,3))
            val /= val.sum(axis=2,keepdims=True)  #make probabilities sum to 1
            distdata = DistData(iid=snpreader.iid,sid=snpreader.sid,pos=snpreader.pos,val=val)
            DistNpz.write(currentFolder + "/../examples/toydata.dist.npz",distdata)
        if True:
            distdata = DistNpz(currentFolder + "/../examples/toydata.dist.npz").read()
            for sid_major,name_bit in [(False,'iidmajor'),(True,'snpmajor')]:
                DistHdf5.write(currentFolder + "/../examples/toydata.{0}.dist.hdf5".format(name_bit),distdata,sid_major=sid_major)
        if True:
            distdata = DistNpz(currentFolder + "/../examples/toydata.dist.npz")[:,:10].read()
            DistNpz.write(currentFolder + "/../examples/toydata10.dist.npz",distdata)
        if True:
            distdata = DistNpz(currentFolder + "/../examples/toydata.dist.npz")[:,:10].read()
            DistMemMap.write(currentFolder + "/../examples/tiny.dist.memmap",distdata)
        print('done')

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()



import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.pstreader import PstReader
from pysnptools.pstreader.pstdata import PstData
import pysnptools.util as pstutil
import warnings

class PstNpz(PstReader):
    '''
    A :class:`.PstReader` for reading \*.pst.npz files from disk.

    See :class:`.PstReader` for general examples of using PstReaders.

    The general NPZ format is described `here <http://docs.scipy.org/doc/numpy/reference/generated/numpy.savez.html>`__. The PstNpz format stores
    val, row, col, row_property, and col_property information in NPZ format.
   
    **Constructor:**
        :Parameters: * **filename** (*string*) -- The PstNpz file to read.

        :Example:

        >>> from pysnptools.pstreader import PstNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> 
        >>> pstnpz_file = example_file('tests/datasets/little.pst.npz')
        >>> data_on_disk = PstNpz(pstnpz_file)
        >>> print(data_on_disk.row_count)
        300

    **Methods beyond** :class:`.NpzReader`

    '''


    def __init__(self, filename):
        '''
        filename    : string of the name of the npz file.
        '''
        super(PstNpz, self).__init__() #We don't send "file name" up because we know about super doesn't want it.
        self._ran_once = False

        self._filename = filename

    def __repr__(self): 
        return "{0}('{1}')".format(self.__class__.__name__,self._filename)

    @property
    def row(self):
        self._run_once()
        return self._row


    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    def _run_once(self):
        if (self._ran_once):
            return
        self._ran_once = True

        with np.load(self._filename,allow_pickle=True) as data: #!! similar code in epistasis
            if len(data.keys()) == 2 and 'arr_0' in data.keys(): #for backwards compatibility
                self._row = data['arr_0']
                self._col = self._row
                self._row_property = np.empty((len(self._row),0))
                self._col_property = np.empty((len(self._col),0))
            else:
                self._row = data['row']
                self._col = data['col']
                if self._row.dtype == self._col.dtype and np.array_equal(self._row, self._col): #If it's square, mark it so by making the col and row the same object
                    self._col = self._row
                self._row_property = data['row_property']
                self._col_property = data['col_property']

        return self

    def copyinputs(self, copier):
        # doesn't need to self._run_once()
        copier.input(self._filename)

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    _read_accepts_slices = True
    def _read(self, row_index_or_none, col_index_or_none, order, dtype, force_python_only, view_ok, num_threads):
        dtype = np.dtype(dtype)
        # 'view_ok' doesn't mean anything here because we are always ready fresh from disk.
        #!! could use mmap so only rows of interest are loaded.
        self._run_once()

        #np.load does the right thing and doesn't load 'val' into memory until accessed here.
        with np.load(self._filename,allow_pickle=True) as data: #!! similar code in epistasis
            if len(data.keys()) == 2 and  'arr_1' in data.keys(): #for backwards compatibility
               val = data['arr_1']
            else:
               val = data['val']

        val, _ = self._apply_sparray_or_slice_to_val(val, row_index_or_none, col_index_or_none, order, dtype, force_python_only, num_threads)
        return val

    @staticmethod
    def write(filename, pstdata):
        """Writes a :class:`PstData` to PstNpz format and returns the :class:`.PstNpz`.

        :param filename: the name of the file to create
        :type filename: string
        :param pstdata: The in-memory data that should be written to disk.
        :type pstdata: :class:`PstData`
        :rtype: :class:`.PstNpz`

        >>> from pysnptools.pstreader import PstData, PstNpz
        >>> import pysnptools.util as pstutil
        >>> data1 = PstData(row=['a','b','c'],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> pstutil.create_directory_if_necessary("tempdir/tiny.pst.npz")
        >>> PstNpz.write("tempdir/tiny.pst.npz",data1)          # Write data in PstNz format
        PstNpz('tempdir/tiny.pst.npz')
        """
        np.savez(filename, row=pstdata.row, col=pstdata.col, row_property=pstdata.row_property, col_property=pstdata.col_property,val=pstdata.val)
        logging.debug("Done writing " + filename)
        return PstNpz(filename)


if __name__ == "__main__":
    import doctest
    logging.basicConfig(level=logging.INFO)
    
    doctest.testmod(optionflags=doctest.ELLIPSIS)

    #from pysnptools.snpreader.dat import Dat
    #snpreader = Dat(r'../tests/datasets/all_chr.maf0.001.N300.dat')
    #snp_matrix = snpreader.read()
    #print(len(snp_matrix['sid']))
    #snp_matrix = snpreader[:,:].read()
    #print(len(snp_matrix['sid']))
    #sid_index_list = snpreader.sid_to_index(['23_9','23_2'])
    #snp_matrix = snpreader[:,sid_index_list].read()
    #print(",".join(snp_matrix['sid']))
    #snp_matrix = snpreader[:,0:10].read()
    #print(",".join(snp_matrix['sid']))

    #print(snpreader.iid_count)
    #print(snpreader.sid_count)
    #print(len(snpreader.pos))

    #snpreader2 = snpreader[::-1,4]
    #print(snpreader.iid_count)
    #print(snpreader2.sid_count)
    #print(len(snpreader2.pos))

    #snp_matrix = snpreader2.read()
    #print(len(snp_matrix['iid']))
    #print(len(snp_matrix['sid']))

    #snp_matrix = snpreader2[5,:].read()
    #print(len(snp_matrix['iid']))
    #print(len(snp_matrix['sid']))

    #iid_index_list = snpreader2.iid_to_index(snpreader2.iid[::2])
    #snp_matrix = snpreader2[iid_index_list,::3].read()
    #print(len(snp_matrix['iid']))
    #print(len(snp_matrix['sid']))

    #snp_matrix = snpreader[[4,5],:].read()
    #print(len(snp_matrix['iid']))
    #print(len(snp_matrix['sid']))

    #print(snpreader2)
    #print(snpreader[::-1,4])
    #print(snpreader2[iid_index_list,::3])
    #print(snpreader[:,sid_index_list])
    #print(snpreader2[5,:])
    #print(snpreader[[4,5],:])



import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.pstreader import PstReader
from pysnptools.pstreader import PstData

class _PstSubset(PstReader):

    def __init__(self, internal, row_indexer, col_indexer):
        '''
        an indexer can be:
             an integer i (same as [i])
             a slice
             a list of integers (including negatives)
             a list of Booleans
        '''
        super(_PstSubset, self).__init__()
        self._ran_once = False

        self._internal = internal
        self._row_indexer = PstReader._make_sparray_or_slice(row_indexer)
        self._col_indexer = PstReader._make_sparray_or_slice(col_indexer)


    def __repr__(self):
        s = "{0}[{1},{2}]".format(self._internal,_PstSubset.static_nice_string(self,self._row_indexer),_PstSubset.static_nice_string(self,self._col_indexer))
        return s

    def copyinputs(self, copier):
        self._internal.copyinputs(copier)

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    _read_accepts_slices = True
    def _read(self, row_indexer, col_indexer, order, dtype, force_python_only, view_ok, num_threads):
        self._run_once()
        dtype = np.dtype(dtype)
        if hasattr(self._internal,'_read_accepts_slices'):
            assert self._internal._read_accepts_slices, "If an object has the _read_accepts_slices attribute, it must have value 'True'"
            composed_row_index_or_none = _PstSubset.compose_indexer_with_indexer(self._internal.row_count, self._row_indexer, self.row_count, row_indexer)
            composed_col_index_or_none = _PstSubset.compose_indexer_with_indexer(self._internal.col_count, self._col_indexer, self.col_count, col_indexer)
            val = self._internal._read(composed_row_index_or_none, composed_col_index_or_none, order, dtype, force_python_only, view_ok, num_threads)
            return val
        else:
            row_index_or_none = PstReader._make_sparray_from_sparray_or_slice(self.row_count, row_indexer)
            composed_row_index_or_none = _PstSubset.compose_indexer_with_index_or_none(self._internal.row_count, self._row_indexer, self.row_count, row_index_or_none)
            col_index_or_none = PstReader._make_sparray_from_sparray_or_slice(self.col_count, col_indexer)
            composed_col_index_or_none = _PstSubset.compose_indexer_with_index_or_none(self._internal.col_count, self._col_indexer, self.col_count, col_index_or_none)
            val = self._internal._read(composed_row_index_or_none, composed_col_index_or_none, order, dtype, force_python_only, view_ok, num_threads)
            return val

    def _run_once(self):
        if self._ran_once:
            return

        self._ran_once = True
        self._row = self._internal.row[self._row_indexer] #!!! would be nice if these four calls were replaced by calls to, e.g. "def _col_index(self,col_indexer)", etc that could be overridden (by for example Pairs)
        self._col = self._internal.col[self._col_indexer]
        if self._row.dtype == self._col.dtype and np.array_equal(self._row,self._col): #When an object is square, keep the row and col the same object.
            self._col = self._row
        self._row_property = self._internal.row_property[self._row_indexer]
        self._col_property = self._internal.col_property[self._col_indexer]

    _slice_format = {(False,False,False):":",
                     (False,False,True):"::{2}",
                     (False,True,False):":{1}",
                     (False,True,True):":{1}:{2}",
                     (True,False,False):"{0}:",
                     (True,False,True):"{0}::{2}",
                     (True,True,False):"{0}:{1}",
                     (True,True,True):"{0}:{1}:{2}"}

    @staticmethod
    def static_nice_string(self, some_slice):
        if isinstance(some_slice,slice):
            return _PstSubset._slice_format[(some_slice.start is not None, some_slice.stop is not None, some_slice.step is not None)].format(some_slice.start, some_slice.stop, some_slice.step)
        elif len(some_slice) == 1:
            return str(some_slice[0])
        elif len(some_slice) < 10:
            return "[{0}]".format(",".join([str(i) for i in some_slice]))
        else:
            return "[{0},...]".format(",".join([str(i) for i in some_slice[:10]]))


    #!!commented out because doesn't guarantee that the shortcut will return with the dtype and order requested.
    #                  Also, didn't handle stacked do-nothing subsets
    #def read(self, order='F', dtype=np.float64, force_python_only=False, view_ok=False):
    #    if view_ok and hasattr(self._internal,"val") and _PstSubset._is_all_slice(self._row_indexer) and _PstSubset._is_all_slice(self._col_indexer):
    #        return self._internal
    #    else:
    #        return PstReader.read(self, order, dtype, force_python_only, view_ok)


    @staticmethod
    def compose_indexer_with_index_or_none(countA, indexerA, countB, index_or_noneB):
        if _PstSubset._is_all_slice(indexerA):
            return index_or_noneB

        indexA = PstReader._make_sparray_from_sparray_or_slice(countA, indexerA)

        if _PstSubset._is_all_slice(index_or_noneB):
            return indexA

        indexAB = indexA[index_or_noneB]

        return indexAB


    @staticmethod
    def compose_indexer_with_indexer(countA, indexerA, countB, indexerB):
        if _PstSubset._is_all_slice(indexerA):
            return indexerB

        if _PstSubset._is_all_slice(indexerB):
            return indexerA

        indexA = PstReader._make_sparray_from_sparray_or_slice(countA, indexerA)
        indexB = PstReader._make_sparray_from_sparray_or_slice(countB, indexerB)

        indexAB = indexA[indexB]

        return indexAB



import numpy as np
import logging
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData
import warnings
from pysnptools.pstreader import PstData
from pysnptools.pstreader import _OneShot

def zero_family(s):
    '''Given an input id from the file, returns 0 as the family id and that input id as the individual id.
    '''
    return "0",s

def zero_pos(s):
    '''Given an input id from the file, return the input id as the sid and returns (0,0,0) as the chromosome, genetic distance, basepair distance.
    '''
    return 0,s,0,0

def just_case(iid):
    '''Use just the individual id (ignoring the family id) in the file.
    '''
    return iid[1]

def just_sid(sid,pos):
    '''Use just the sid (ignoring the chromosome, genetic distance, and basepair distance) in the file.
    '''
    return sid

class Dense(_OneShot,SnpReader):
    '''
    A :class:`.SnpReader` for reading \*.dense.txt (0,1,2 text files) from disk.

    See :class:`.SnpReader` for general examples of using SnpReaders.

    This format looks like::

        var	4006	9570    22899	37676	41236	41978	55159	66828...
        1-10004-rs12354060	22222222...
        1-707348-rs12184279	222222?2...
        1-724325-rs12564807	00000000...
        ...

    where rows are sids and columns are iids.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The Dense file to read.
                     * **extract_iid_function** (*string*) -- A function that breaks the row id from the file into a family id and an individual id. Defaults to
                       setting the family id to 0 and using the whole row id as the iid.
                     * **extract_sid_pos_function** (*string*) -- A function that breaks the column id from the file into (chromosome, sid, genetic distance, basepair distance).
                       Defaults to setting the :attr:`SnpReader.pos` information to 0 and using the whole column id as the sid.

        :Example:

        >>> from pysnptools.snpreader import Dense
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> dense_file = example_file("pysnptools/examples/toydata100.dense.txt")
        >>> data_on_disk = Dense(dense_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 100)

    **Methods beyond** :class:`.SnpReader`

    '''
    def __init__(self, filename, extract_iid_function=zero_family, extract_sid_pos_function=zero_pos):
        '''
        filename    : string of the name of the Dat file.
        '''
        super(Dense, self).__init__()
        self.filename = filename
        self.extract_iid_function = extract_iid_function
        self.extract_sid_pos_function = extract_sid_pos_function

    def _read_pstdata(self):
        bim_list = []
        val_list_list = []
        with open(self.filename,"r") as fp:
            header = fp.readline()
            iid_string_list = header.strip().split()[1:]
            iid = np.array([self.extract_iid_function(iid_string) for iid_string in iid_string_list],dtype='str')
            val_list = []
            zerofloat = float('0'[0])
            missing_char = "?"[0]
            for line_index,line in enumerate(fp):
                if line_index % 1000 == 0:
                    logging.info("reading sid and iid info from line {0} of file '{1}'".format(line_index, self.filename))
                sid_string_rest = line.strip().split()
                sid_string = sid_string_rest[0]
                rest = [] if len(sid_string_rest)==1 else sid_string_rest[1]
                assert len(rest) == len(iid)
                bim_list.append(self.extract_sid_pos_function(sid_string))
                val_list = np.array([float(val)-zerofloat if val!=missing_char else np.NaN for val in rest])
                val_list_list.append(val_list)

        col = np.array([bim[1] for bim in bim_list],dtype='str')
        col_property = np.array([[bim[0],bim[2],bim[3]] for bim in bim_list],dtype=np.float64)

        val = np.zeros((len(iid),len(col)))
        for col_index in range(len(col)):
            val[:,col_index] = val_list_list[col_index]

        return PstData(iid,col,val,col_property=col_property,name=self.filename)

    @staticmethod
    def write(filename, snpdata, join_iid_function=just_case,join_sid_pos_function=just_sid):
        """Writes a :class:`SnpData` to Dense format. The values must be 0,1,2 (or missing). Returns a :class:`.Dense` with the default extractors.

        :param filename: the name of the file to create
        :type filename: string
        :param snpdata: The in-memory data that should be written to disk.
        :type snpdata: :class:`SnpData`
        :param join_iid_function: function to turn a family id and individual id into a file id for columns.
                                  Defaults ignoring the family id and using the individual id as the column id.
        :type join_iid_function: a function
        :param join_sid_pos_function: function to turn a sid and pos data into a file id for rows.
                                      Defaults ignoring the :attr:`SnpReader.pos` information and using the sid id as the row id.
        :type join_sid_pos_function: a function
        :rtype: :class:`.Dense`

        >>> from pysnptools.snpreader import Dense, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpdata = Bed(bed_file,count_A1=False)[:,:10].read()  # Read first 10 snps from Bed format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.dense.txt")
        >>> Dense.write("tempdir/toydata10.dense.txt",snpdata)        # Write data in Dense format
        Dense('tempdir/toydata10.dense.txt')
        """

        if isinstance(filename,SnpData) and isinstance(snpdata,'S'): #For backwards compatibility, reverse inputs if necessary
            warnings.warn("write statement should have filename before data to write", DeprecationWarning)
            filename, snpdata = snpdata, filename 

        snpsarray = snpdata.val
        with open(filename,"w") as filepointer:
            filepointer.write("var\t")
            filepointer.write("\t".join((join_iid_function(iid_pair) for iid_pair in snpdata.iid)) + "\n")

            for sid_index, sid in enumerate(snpdata.sid):
                pos = snpdata.pos[sid_index]
                if sid_index % 1000 == 0:
                    logging.info("Writing snp # {0} to file '{1}'".format(sid_index, filename))
                filepointer.write("%s\t" % join_sid_pos_function(sid,pos))
                row = snpsarray[:,sid_index]
                filepointer.write("".join((str(int(i)) if i==i else "?" for i in row)) + "\n")
        logging.info("Done writing " + filename)
        return Dense(filename)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
import logging
import doctest
import unittest
import os.path
import time

from pysnptools.kernelreader import *
from pysnptools.snpreader import Bed
from pysnptools.util import create_directory_if_necessary
from pysnptools.pstreader import PstReader
from pysnptools.snpreader import SnpData
import pysnptools.standardizer as stdizer

class _fortesting_JustCheckExists(object): #Implements ICopier

    def __init__(self,doPrintOutputNames=False):
        self.doPrintOutputNames = doPrintOutputNames
    
    def input(self,item):
        if isinstance(item, str):
            if not os.path.exists(item): raise Exception("Missing input file '{0}'".format(item))
        elif hasattr(item,"copyinputs"):
            item.copyinputs(self)
        # else -- do nothing

    def output(self,item):
        if isinstance(item, str):
            if not os.path.exists(item): raise Exception("Missing output file '{0}'".format(item))
            if self.doPrintOutputNames:
                print(item)
        elif hasattr(item,"copyoutputs"):
            item.copyoutputs(self)
        # else -- do nothing




class TestKernelReader(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.currentFolder = os.path.dirname(os.path.realpath(__file__))

    def test_merge_std(self):
        #unit vs beta
        for std in [stdizer.Beta(2,10),stdizer.Unit()]:
            np.random.seed(0)
            sid_count = 20
            snpreader = SnpData(iid=[["0","0"],["1","1"],["2","2"]],sid=[str(i) for i in range(sid_count)],val=np.array(np.random.randint(3,size=[3,sid_count]),dtype=np.float64,order='F'))
            kerneldata0,trained0,diag0 = SnpKernel(snpreader,std,block_size=1)._read_with_standardizing(to_kerneldata=True,return_trained=True)
            kerneldata1,trained1,diag1 = SnpKernel(snpreader,std,block_size=None)._read_with_standardizing(to_kerneldata=True,return_trained=True)
            np.testing.assert_array_almost_equal(kerneldata0.val,kerneldata1.val, decimal=10)
            np.testing.assert_array_almost_equal(trained0.stats,trained1.stats, decimal=10)
            assert abs(diag0.factor-diag1.factor) < 1e-7
    
    def test_cpp_std(self):

        #Order C vs F
        for order in ['C','F']:
            #32 vs 64
            for dtype in [np.float64,np.float32]:
                #unit vs beta
                for std in [stdizer.Unit(),stdizer.Beta(2,10)]:
                        np.random.seed(0)
                        sid_count = 20
                        snpreader0 = SnpData(iid=[["0","0"],["1","1"],["2","2"]],sid=[str(i) for i in range(sid_count)],val=np.array(np.random.randint(3,size=[3,sid_count]),dtype=dtype,order=order))
                        snpreader1 = SnpData(iid=[["3","3"],["4","4"]],sid=[str(i) for i in range(sid_count)],val=np.array(np.random.randint(3,size=[2,sid_count]),dtype=dtype,order=order))

                        #has SNC
                        for has_SNC_in_train in [False, True]:
                            if has_SNC_in_train:
                                snpreader0.val[:,1] = 0

                            #missing data
                            for has_missing_data in [False, True]:
                                if has_missing_data:
                                    snpreader0.val[0,2]=np.nan
                                    snpreader1.val[0,2]=np.nan

                                #gather stats vs not
                                cppa, stdcppa = snpreader0.read(order=order,dtype=dtype).standardize(std,return_trained=True,force_python_only=False)
                                pya, stdpya = snpreader0.read(order=order,dtype=dtype).standardize(std,return_trained=True,force_python_only=True)
                                np.testing.assert_array_almost_equal(cppa.val, pya.val, decimal=10 if dtype==np.float64 else 5)

                                np.testing.assert_array_almost_equal(stdcppa.stats,stdpya.stats, decimal=10 if dtype==np.float64 else 5)
                                assert (np.inf in stdcppa.stats[:,1]) == has_SNC_in_train
                                assert (np.inf in stdpya.stats[:,1]) == has_SNC_in_train

                                if has_SNC_in_train:
                                    assert np.array_equal(cppa.val[:,1],np.zeros([cppa.val.shape[0]]))
                                    assert np.array_equal(pya.val[:,1],np.zeros([pya.val.shape[0]]))

                                if has_missing_data:
                                    assert 0 == cppa.val[0,2]
                                    assert 0 == pya.val[0,2]
                                        
                                #uses stats
                                cppb = snpreader1.read(order=order,dtype=dtype).standardize(stdcppa,force_python_only=False)
                                pyb = snpreader1.read(order=order,dtype=dtype).standardize(stdpya,force_python_only=True)
                                np.testing.assert_array_almost_equal(cppb.val, pyb.val, decimal=10 if dtype==np.float64 else 5)
                                np.testing.assert_array_almost_equal(stdcppa.stats,stdpya.stats, decimal=10 if dtype==np.float64 else 5) #Make sure we haven't messed up the train stats

                                if has_SNC_in_train:
                                    assert np.array_equal(cppb.val[:,1],np.zeros([cppb.val.shape[0]]))
                                    assert np.array_equal(pyb.val[:,1],np.zeros([pyb.val.shape[0]]))

                                if has_missing_data:
                                    assert cppb.val[0,2]==0
                                    assert pyb.val[0,2]==0
        logging.info("done with 'test_cpp_std'")


    def test_intersection(self):

        from pysnptools.standardizer import Unit
        from pysnptools.kernelreader import SnpKernel
        from pysnptools.snpreader import Pheno
        from pysnptools.kernelreader._subset import _KernelSubset
        from pysnptools.snpreader._subset import _SnpSubset
        from pysnptools.util import intersect_apply

        snps_all = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        k = SnpKernel(snps_all,stdizer.Identity())

        pheno = Pheno(self.currentFolder + "/../examples/toydata.phe")
        pheno = pheno[1:,:] # To test intersection we remove a iid from pheno

        k1,pheno = intersect_apply([k,pheno]) #SnpKernel is special because it standardizes AFTER intersecting.
        assert isinstance(k1.snpreader,_SnpSubset) and not isinstance(k1,_KernelSubset)

        #What happens with fancy selection?
        k2 = k[::2]
        assert isinstance(k2,SnpKernel)

        logging.info("Done with test_intersection")

    def test_respect_read_inputs(self):
        from pysnptools.kernelreader import KernelHdf5,Identity,KernelNpz,SnpKernel
        from pysnptools.standardizer import Unit
        from pysnptools.standardizer import Identity as StdIdentity
        from pysnptools.snpreader import Bed

        previous_wd = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        iidref = KernelNpz('../examples/toydata.kernel.npz').iid

        for kernelreader in [
                           SnpKernel(Bed('../examples/toydata.5chrom.bed',count_A1=True),StdIdentity())[::2,::2],
                           Bed('../examples/toydata.5chrom.bed',count_A1=True)[::2,::2].read_kernel(StdIdentity()),
                           KernelHdf5('../examples/toydata.kernel.hdf5'),
                           Identity(iidref,test=[('0','x'),('0','y')]),
                           Identity(iidref),
                           KernelNpz('../examples/toydata.kernel.npz'),
                           KernelNpz('../examples/toydata.kernel.npz').read(),
                           KernelNpz('../examples/toydata.kernel.npz')[::2,::2],
                           Bed('../examples/toydata.5chrom.bed',count_A1=True).read_kernel(Unit()),
                           SnpKernel(Bed('../examples/toydata.5chrom.bed',count_A1=True),Unit())
                          ]:
            logging.info(str(kernelreader))
            for order in ['F','C','A']:
                for dtype in [np.float32,np.float64]:
                    for force_python_only in [True,False]:
                        for view_ok in [True,False]:
                            val = kernelreader.read(order=order,dtype=dtype,force_python_only=force_python_only,view_ok=view_ok).val
                            has_right_order = order=="A" or (order=="C" and val.flags["C_CONTIGUOUS"]) or (order=="F" and val.flags["F_CONTIGUOUS"])
                            if hasattr(kernelreader,'val') and not view_ok:
                                assert kernelreader.val is not val
                            if (hasattr(kernelreader,'val') and view_ok and kernelreader.val is not val and
                                (order == 'A' or (order == 'F' and kernelreader.val.flags['F_CONTIGUOUS']) or (order == 'C' and kernelreader.val.flags['C_CONTIGUOUS'])) and
                                (dtype is None or  kernelreader.val.dtype == dtype)):
                                logging.info("{0} could have read a view, but didn't".format(distreader))
                            assert val.dtype == dtype and has_right_order
        os.chdir(previous_wd)


    def test_respect_inputs(self):
        np.random.seed(0)
        for dtype_start,decimal_start in [(np.float32,5),(np.float64,10)]:
            for order_start in ['F','C','A']:
                for sid_count in [20,2]:
                    snpdataX = SnpData(iid=[["0","0"],["1","1"],["2","2"]],sid=[str(i) for i in range(sid_count)],val=np.array(np.random.randint(3,size=[3,sid_count]),dtype=dtype_start,order=order_start))
                    for stdx in [stdizer.Beta(1,25),stdizer.Identity(),stdizer.Unit()]:
                        for snpreader0 in [snpdataX,snpdataX[:,1:]]:
                            snpreader1 = snpreader0[1:,:]

                            refdata0, trained_standardizer = snpreader0.read().standardize(stdx,return_trained=True,force_python_only=True)
                            refval0 = refdata0.val.dot(refdata0.val.T)
                            refdata1 = snpreader1.read().standardize(trained_standardizer,force_python_only=True)#LATER why aren't these used?
                            refval1 = refdata0.val.dot(refdata1.val.T)#LATER why aren't these used?
                            for dtype_goal,decimal_goal in [(np.float32,5),(np.float64,10)]:
                                for order_goal in ['F','C','A']:
                                    k = snpreader0.read_kernel(standardizer=stdx,block_size=1,order=order_goal,dtype=dtype_goal)
                                    PstReader._array_properties_are_ok(k.val,order_goal,dtype_goal)
                                    np.testing.assert_array_almost_equal(refval0,k.val, decimal=min(decimal_start,decimal_goal))

    def test_fail(self):
        did_fail = True
        try:
            kd = KernelData(iid=[["0"],["1"],["2"]],val=[[1,2,3],[4,5,6],[7,8,9]]) #Wrong iid shape
            did_fail = False
        except:
            pass
        assert did_fail, "The constructor should fail because the iid is the wrong shape"

    def test_kernel2(self):
        logging.info("in kernel2")
        kd = KernelData(iid=[["0","0"],["1","1"],["2","2"]],val=[[1,2,3],[4,5,6],[7,8,9]])
        assert np.array_equal(kd.iid_to_index([["1","1"],["2","2"]]),np.array([1,2]))
        assert np.array_equal(kd.iid0_to_index([["1","1"],["2","2"]]),np.array([1,2]))
        kd = kd.standardize()
        assert np.abs(np.diag(kd.val).sum()-3)<1e-7
        assert kd.iid1_count == 3

    def test_snp_kernel2(self):
        logging.info("in test_snp_kernel2")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        snpkernel = SnpKernel(snpreader,standardizer=stdizer.Beta(1,25))
        s  = str(snpkernel)
        _fortesting_JustCheckExists().input(snpkernel)
        
    def test_npz(self):
        logging.info("in test_npz")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        kerneldata1 = snpreader.read_kernel(standardizer=stdizer.Unit())
        s = str(kerneldata1)
        output = "tempdir/kernelreader/toydata.kernel.npz"
        create_directory_if_necessary(output)
        KernelNpz.write(output,kerneldata1)
        kernelreader2 = KernelNpz(output)
        kerneldata2 = kernelreader2.read()
        np.testing.assert_array_almost_equal(kerneldata1.val, kerneldata2.val, decimal=10)
        logging.info("done with test")

    def test_subset(self):
        logging.info("in test_subset")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        snpkernel = SnpKernel(snpreader,stdizer.Unit())
        krsub = snpkernel[::2,::2]
        kerneldata1 = krsub.read()
        expected = snpreader.read_kernel(stdizer.Unit())[::2].read()
        np.testing.assert_array_almost_equal(kerneldata1.val, expected.val, decimal=10)

        krsub2 = snpkernel[::2]
        kerneldata2 = krsub2.read()
        np.testing.assert_array_almost_equal(kerneldata2.val, expected.val, decimal=10)
        logging.info("done with test")

    def test_identity(self):
        logging.info("in test_identity")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        assert snpreader.iid is snpreader.row
        kid = Identity(snpreader.row)
        assert np.array_equal(kid.row,kid.iid) and np.array_equal(kid.iid,kid.iid0) and np.array_equal(kid.iid0,kid.iid1) and np.array_equal(kid.iid1, kid.col)
        np.testing.assert_array_almost_equal(kid.read().val, np.identity(snpreader.iid_count))

        #subset
        sub1 = kid[1:5]
        np.testing.assert_array_almost_equal(sub1.read().val, np.identity(4))

        #zero size
        sub2 = kid[0:0]
        np.testing.assert_array_almost_equal(sub2.read().val, np.identity(0))

    def test_identity_sub(self):
        logging.info("in test_identity_sub")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        assert snpreader.iid is snpreader.row
        kid = Identity(snpreader.row)
        sub3 = kid[::2,1:5]
        expected = np.identity(kid.iid_count)[::2,:][:,1:5]
        np.testing.assert_array_almost_equal(sub3.read().val,expected)

        logging.info("done with test")

    def test_underscore_read1(self):
        logging.info("in test_underscore_read1")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        assert snpreader.iid is snpreader.row
        kid = Identity(snpreader.row)
        sub3 = kid[::2,::2]
        expected = np.identity(kid.iid_count)[::2,:][:,::2]
        np.testing.assert_array_almost_equal(sub3.read().val,expected)

        logging.info("done with test")

    def test_underscore_read2(self):
        logging.info("in test_underscore_read2")
        snpreader = Bed(self.currentFolder + "/../examples/toydata.5chrom.bed",count_A1=False)
        assert snpreader.iid is snpreader.row
        kid = Identity(snpreader.row)
        sub3 = kid[::2,::2]
        expected = np.identity(kid.iid_count)[::2,:][:,::2]
        np.testing.assert_array_almost_equal(sub3.read().val,expected)

        logging.info("done with test")

# We do it this way instead of using doctest.DocTestSuite because doctest.DocTestSuite requires modules to be pickled, which python doesn't allow.
# We need tests to be pickleable so that they can be run on a cluster.
class TestKrDocStrings(unittest.TestCase):
    def test_kernelreader(self):
        import pysnptools.kernelreader.kernelreader
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.kernelreader.kernelreader,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_snpkernel(self):
        import pysnptools.kernelreader.snpkernel
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.kernelreader.snpkernel,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_kerneldata(self):
        import pysnptools.kernelreader.kerneldata
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.kernelreader.kerneldata,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_identity(self):
        import pysnptools.kernelreader.identity
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.kernelreader.identity,optionflags=doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_kernelhdf5(self):
        import pysnptools.kernelreader.kernelhdf5
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.kernelreader.kernelhdf5,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__

    def test_kernelnpz(self):
        import pysnptools.kernelreader.kernelnpz
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(pysnptools.kernelreader.kernelnpz,optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__



def getTestSuite():
    """
    set up composite test suite
    """
    
    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestKernelReader))
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestKrDocStrings))
    return test_suite

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()




import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.kernelreader import KernelReader
from pysnptools.pstreader import PstData
from pysnptools.pstreader import PstReader

class Identity(KernelReader):
    '''
    A :class:`.KernelReader` for that represents an identity matrix. No memory for the values is allocated until :meth:`Identity.read` is called.

    See :class:`.KernelReader` for general examples of using KernelReaders.

    **Constructor:**
        :Parameters: * **iid** (an array of strings) -- The :attr:`KernelReader.iid` information

        :Example:

        >>> from pysnptools.kernelreader import Identity
        >>> identity = Identity(iid=[['fam0','iid0'],['fam0','iid1']])
        >>> print(identity.iid_count)
        2
        >>> print(identity.read().val) # '...' is possible space character
        [[...1.  0.]
         [...0.  1.]]

        >>> identity = Identity(iid=[['fam0','iid0'],['fam0','iid1'],['fam0','iid2']],test=[['fam0','iid1'],['fam0','iid3']])
        >>> print((identity.iid0_count, identity.iid1_count))
        (3, 2)
        >>> print(identity.read().val) # '...' is possible space character
        [[...0.  0.]
         [...1.  0.]
         [...0.  0.]]

    '''
    def __init__(self, iid, test=None): #!!! add docs and test for test
        super(Identity, self).__init__()

        if test is None:
            test = iid

        if test is iid:
            iid = PstData._fixup_input(iid,empty_creator=lambda ignore:np.empty([0,2],dtype='str'),dtype='str')
            test = iid
        else:
            iid = PstData._fixup_input(iid,empty_creator=lambda ignore:np.empty([0,2],dtype='str'),dtype='str')
            test = PstData._fixup_input(test,empty_creator=lambda ignore:np.empty([0,2],dtype='str'),dtype='str')

        if len(iid)>0:
            self._row0 = iid
        else:
            self._row0 = self._empty

        if len(test)>0:
            self._row1 = test
        else:
            self._row1 = self._empty

    _empty = np.empty([0,2],dtype='str')

    def __repr__(self): 
        return "{0}({1}x{2})".format(self.__class__.__name__, self.row_count, self.col_count)


    @property
    def row(self):
        return self._row0

    @property
    def col(self):
        return self._row1

    def _read(self, row_index_or_none, col_index_or_none, order, dtype, force_python_only, view_ok, num_threads):
        dtype = np.dtype(dtype)
        if row_index_or_none is None and col_index_or_none is None and self._row0 is self._row1: #read all of a square ID
            val = np.identity(self.row_count,dtype=dtype)
            if (order=='F'and not val.flags["F_CONTIGUOUS"]) or (order=='C'and not val.flags["C_CONTIGUOUS"]):
                val = val.T
            return val
        else: #Non-square
            #!!! This is also less efficient than it could be because it create a big identity matrix and then slices it.

            #In about O(col_count + row_count) fill in zeros
            big = np.zeros([self.row_count,self.col_count],dtype=dtype)
            common = set([PstReader._makekey(x) for x in self.row]) & set([PstReader._makekey(x) for x in self.col])
            big[self.row_to_index(common),self.col_to_index(common)] = 1.0
            val, shares_memory = self._apply_sparray_or_slice_to_val(big, row_index_or_none, col_index_or_none, order, dtype, force_python_only, num_threads)
            return val

    def __getitem__(self, iid_indexer_and_snp_indexer):
        if isinstance(iid_indexer_and_snp_indexer,tuple):
            row_index_or_none, col_index_or_none = iid_indexer_and_snp_indexer
        else:
            row_index_or_none = iid_indexer_and_snp_indexer
            col_index_or_none = row_index_or_none

        #If iid0 will be iid1 then make 'is' equal
        iid=self.iid0[row_index_or_none]
        test=self.iid1[col_index_or_none]
        if len(iid) == len(test) and np.array_equal(iid,test):
            result = Identity(iid)
        else: 
            result = Identity(iid=iid,test=test)
        return result


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE)



import numpy as np
from itertools import *
import pandas as pd
import logging
import time
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData


class _Dist2Snp(SnpReader):
    def __init__(self, snpreader, max_weight=2.0, block_size=None):
        super(_Dist2Snp, self).__init__()

        self.distreader = snpreader
        self.max_weight=max_weight
        self.block_size = block_size

    @property
    def row(self):
        return self.distreader.iid

    @property
    def col(self):
        return self.distreader.col

    def __repr__(self):
        return self._internal_repr()

    def _internal_repr(self): #!!! merge this with __repr__
        s = "{0}.as_snp(".format(self.distreader)
        if self.block_size is not None:
            s += "block_size={0}".format(self.block_size)
        s += ")"
        return s

    def copyinputs(self, copier):
        #Doesn't need run_once
        copier.input(self.distreader)

    def _read(self, row_index_or_none, col_index_or_none, order, dtype, force_python_only, view_ok, num_threads):
        from pysnptools.distreader import DistReader
        dtype = np.dtype(dtype)


        assert row_index_or_none is None and col_index_or_none is None #real assert because indexing should already be pushed to the inner distreader

        weights = np.array([0,.5,1],dtype=dtype)*self.max_weight

        #Do all-at-once (not in blocks) if 1. No block size is given or 2. The #ofSNPs < Min(block_size,iid_count)
        if self.block_size is None or (self.sid_count <= self.block_size or self.sid_count <= self.iid_count):
            distdata = DistReader._as_distdata(self.distreader,dtype=dtype,order=order,force_python_only=force_python_only,num_threads=num_threads)
            val = (distdata.val*weights).sum(axis=-1)
            has_right_order = order="A" or (order=="C" and val.flags["C_CONTIGUOUS"]) or (order=="F" and val.flags["F_CONTIGUOUS"])
            assert has_right_order
            return val
        else: #Do in blocks
            t0 = time.time()
            if order=='A':
                order = 'F'
            val = np.zeros([self.iid_count,self.sid_count],dtype=dtype,order=order)#LATER should use empty or fillnan

            logging.info("reading {0} distribution data in blocks of {1} SNPs and finding expected values (for {2} individuals)".format(self.sid_count, self.block_size, self.iid_count))
            ct = 0
            ts = time.time()

            for start in range(0, self.sid_count, self.block_size):
                ct += self.block_size
                distdata = self.distreader[:,start:start+self.block_size].read(order=order,dtype=dtype,force_python_only=force_python_only,view_ok=True,num_threads=num_threads) # a view is always OK, because we'll allocate memory in the next step
                val[:,start:start+self.block_size] = (distdata.val*weights).sum(axis=-1)
                if ct % self.block_size==0:
                    diff = time.time()-ts
                    if diff > 5:
                        logging.info("read %s SNPs in %.2f seconds" % (ct, diff))

            t1 = time.time()
            logging.info("%.2f seconds elapsed" % (t1-t0))

            return val

    def __getitem__(self, iid_indexer_and_snp_indexer):
        row_index_or_none, col_index_or_none = iid_indexer_and_snp_indexer
        return _Dist2Snp(self.distreader[row_index_or_none,col_index_or_none],max_weight=self.max_weight,block_size=self.block_size)


    @property
    def sid(self):
        '''The :attr:`SnpReader.sid` property of the SNP data.
        '''
        return self.distreader.sid

    @property
    def sid_count(self):
        '''The :attr:`SnpReader.sid_count` property of the SNP data.
        '''
        return self.distreader.sid_count

    @property
    def pos(self):
        '''The :attr:`SnpReader.pos` property of the SNP data.
        '''
        return self.distreader.pos


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod()
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import os
import numpy as np
import logging
import unittest
import doctest
from pathlib import Path
from pysnptools.util.pairs import _Pairs as UtilPairs

from pysnptools.snpreader import SnpReader

class _Pairs(SnpReader):
    '''
    Experimental.
    '''

    #!!!see fastlmm\association\epistasis.py for code that allows ranges of snps to be specified when making pairs
    def __init__(self, snpreader0,snpreader1=None, do_standardize=True,sid_materialize_limit=1000*1000,_include_single_times_single=False): 
        #!!! could add option to change snp separator and another to encode chrom, etc in the snp name
        super(_Pairs, self).__init__()
        self._ran_once = False
        self.snpreader0 = snpreader0
        self.snpreader1 = snpreader1
        self.do_standardize = do_standardize
        self.sid_materialize_limit = sid_materialize_limit
        self._include_single_times_single=_include_single_times_single

    def __repr__(self):
        part2 = '' if self.snpreader1 is None else ',{0}'.format(self.snpreader1)
        return "{0}({1}{2})".format(self.__class__.__name__,self.snpreader0,part2)

    @property
    def row(self):
        """*same as* :attr:`iid`
        """
        if not hasattr(self,"_row"):
            self._row = self.snpreader0.row
            assert self.snpreader1 is None or np.array_equal(self._row,self.snpreader1.row), "Expect snpreaders to have the same iids in the same order"
        return self._row

    @property
    def col(self):
        """*same as* :attr:`sid`
        """
        if not hasattr(self,"_col"):
            assert self.col_count < self.sid_materialize_limit, '{:,} is too many sids to materialize'.format(self.col_count)
            if self.snpreader1 is None:
                col_0 = self.snpreader0.col
                col_list = []
                self.index0_list = [] #!!!should be _index0_list, etc
                self.index1_list = []
                for index0 in xrange(self.snpreader0.col_count):
                    #logging.info("index0={0} of {1}".format(index0,self.snpreader0.col_count))#!!!
                    start1 = index0 if self._include_single_times_single else index0+1
                    for index1 in xrange(start1,self.snpreader0.col_count):
                        col_list.append('{0},{1}'.format(col_0[index0],col_0[index1]))
                        self.index0_list.append(index0)
                        self.index1_list.append(index1)
                self._col = np.array(col_list)
                self.index0_list = np.array(self.index0_list)
                self.index1_list = np.array(self.index1_list)
            else:
                col_0 = self.snpreader0.col
                col_1 = self.snpreader1.col
                assert len(set(col_0)&set(col_1))==0,"Pairs currently requires two snpreaders to have no sids in common"
                col_list = []
                self.index0_list = []
                self.index1_list = []
                for index0 in xrange(self.snpreader0.col_count):
                    #logging.info("index0={0} of {1}".format(index0,self.snpreader0.col_count))#!!!
                    for index1 in xrange(self.snpreader1.col_count):
                        col_list.append('{0},{1}'.format(col_0[index0],col_1[index1]))
                        self.index0_list.append(index0)
                        self.index1_list.append(index1)
                self._col = np.array(col_list)
                self.index0_list = np.array(self.index0_list)
                self.index1_list = np.array(self.index1_list)

        assert self.col_count == len(self._col), "real assert"
        return self._col

    @property
    def col_count(self):
        n0 = self.snpreader0.col_count
        if self.snpreader1 is None:
            if self._include_single_times_single:
                return (n0*n0+n0)//2
            else:
                return (n0*n0-n0)//2
        else:
            return n0*self.snpreader1.col_count

    @property
    def col_property(self):
        """*same as* :attr:`pos`
        """
        if not hasattr(self,"_col_property"):
            self._col_property = np.zeros([self.sid_count,3],dtype=np.int64)
        return self._col_property

    def copyinputs(self, copier):
        # doesn't need to self._run_once() because only uses original inputs !!!is this true?
        self.snpreader0.copyinputs(copier)
        if self.snpreader1 is not None:
            self.snpreader1.copyinputs(copier)

    def _run_once(self):
        if self._ran_once:
            return

        self._ran_once = True
        self.col

    def _read(self, iid_index_or_none, sid_index_or_none, order, dtype, force_python_only, view_ok, num_threads):
        self._run_once()
        dtype = np.dtype(dtype)

        iid_count_in = self.iid_count
        sid_count_in = self.sid_count

        if iid_index_or_none is not None:
            iid_count_out = len(iid_index_or_none)
            iid_index = iid_index_or_none
        else:
            iid_count_out = iid_count_in
            iid_index = range(iid_count_in)

        if sid_index_or_none is not None:
            sid_count_out = len(sid_index_or_none)
            sid_index = sid_index_or_none
        else:
            sid_count_out = sid_count_in
            sid_index = range(sid_count_in)

        
        sid_index_inner_0 = self.index0_list[sid_index] #Find the sid_index of the left snps of interest
        sid_index_inner_1 = self.index1_list[sid_index] #Find the sid_index of the right snps of interest
        if self.snpreader1 is None:
            sid_index_inner_01 = np.unique(np.r_[sid_index_inner_0,sid_index_inner_1]) #Index of every snp of interest
            inner_01 = self.snpreader0[iid_index_or_none,sid_index_inner_01].read(order=order, dtype=dtype, force_python_only=force_python_only, view_ok=True,num_threads=num_threads) #read every val of interest
            val_inner_01 = inner_01.standardize(num_threads=num_threads).val if self.do_standardize else inner_01.val

            sid_index_inner_01_reverse = {v:i for i,v in enumerate(sid_index_inner_01)} #Dictionary of snp_index to position in sid_index_inner_01
            sid_index_inner_0_in_val = np.array([sid_index_inner_01_reverse[i] for i in sid_index_inner_0])  #Replace snp_index0 with column # in val_inner_01
            sid_index_inner_1_in_val = np.array([sid_index_inner_01_reverse[i] for i in sid_index_inner_1])  #Replace snp_index1 with column # in val_inner_01
            val_inner_0 = val_inner_01[:,sid_index_inner_0_in_val] #Extract the vals for the left snps of interest
            val_inner_1 = val_inner_01[:,sid_index_inner_1_in_val]#Extract the vals for the right snps of interest
        else:
            inner_0 = self.snpreader0[iid_index_or_none,sid_index_inner_0].read(order=order, dtype=dtype, force_python_only=force_python_only, view_ok=True,num_threads=num_threads) #read every val of interest
            inner_1 = self.snpreader1[iid_index_or_none,sid_index_inner_1].read(order=order, dtype=dtype, force_python_only=force_python_only, view_ok=True,num_threads=num_threads) #read every val of interest
            val_inner_0 = inner_0.standardize(num_threads=num_threads).val if self.do_standardize else inner_0.val
            val_inner_1 = inner_1.standardize(num_threads=num_threads).val if self.do_standardize else inner_1.val
        val = val_inner_0*val_inner_1 #Element multiplication creates the vals for the pairs
        return val



def split_on_sids(snpreader,part_count):
    sid_count = snpreader.sid_count
    start = 0
    for part_index in xrange(1,part_count+1):
        end = part_index*sid_count//part_count
        yield snpreader[:,start:end]
        start=end


class _Pairs2(SnpReader):
    
    def __init__(self, snpreader0,snpreader1=None, do_standardize=True,sid_materialize_limit=1000*1000,_include_single_times_single=False): #!!! could add option to change snp separator and another to encode chrom, etc in the snp name
        super(_Pairs2, self).__init__()
        self._ran_once = False
        self.snpreader0 = snpreader0
        self.snpreader1 = snpreader1
        self.do_standardize = do_standardize
        self.sid_materialize_limit = sid_materialize_limit
        self._include_single_times_single=_include_single_times_single
        self._utilpairs = UtilPairs(snpreader0.sid,snpreader1.sid if snpreader1 is not None else snpreader0.sid,include_singles=_include_single_times_single,duplicates_ok=False)

    def __repr__(self):
        part2 = '' if self.snpreader1 is None else ',{0}'.format(self.snpreader1)
        return "{0}({1}{2})".format(self.__class__.__name__,self.snpreader0,part2)

    @property
    def row(self):
        """*same as* :attr:`iid`
        """
        if not hasattr(self,"_row"):
            self._row = self.snpreader0.row
            assert self.snpreader1 is None or np.array_equal(self._row,self.snpreader1.row), "Expect snpreaders to have the same iids in the same order"
        return self._row

    @property
    def col(self):
        """*same as* :attr:`sid`
        """
        if not hasattr(self,"_col"):
            assert self.col_count < self.sid_materialize_limit, '{:,} is too many sids to materialize'.format(self.col_count)
            snpreader1 = self.snpreader1 if self.snpreader1 is not None else self.snpreader0
            #!!!self.index0_list = self.snpreader0.sid_to_index((sid0 for sid0,sid1 in self._utilpairs[:])) #!!! can we do without these?
            #!!!self.index1_list = snpreader1.sid_to_index(sid1 for sid0,sid1 in self._utilpairs[:])#!!!can we do without these?
            self._col = np.array(list(','.join(pair) for pair in self._utilpairs[:]))
        return self._col

    @property
    def col_count(self):
        return len(self._utilpairs)

    @property
    def col_property(self):
        """*same as* :attr:`pos`
        """
        if not hasattr(self,"_col_property"):
            self._col_property = np.zeros([self.sid_count,3],dtype=np.int64)
        return self._col_property

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because only uses original inputs !!!is this true?
        self.snpreader0.copyinputs(copier)
        if self.snpreader1 is not None:
            self.snpreader1.copyinputs(copier)

    def run_once(self):
        if self._ran_once:
            return

        self._ran_once = True
        self.col

    def _read(self, iid_index_or_none, sid_index_or_none, order, dtype, force_python_only, view_ok, num_threads):
        self.run_once()

        iid_count_in = self.iid_count
        sid_count_in = self.sid_count

        if iid_index_or_none is not None:
            iid_count_out = len(iid_index_or_none)
            iid_index_out = iid_index_or_none
        else:
            iid_count_out = iid_count_in
            iid_index_out = range(iid_count_in)

        if sid_index_or_none is not None:
            sid_count_out = len(sid_index_or_none)
            sid_index_out = sid_index_or_none
        else:
            sid_count_out = sid_count_in
            sid_index_out = splice(None)#!!! test this

        pair_array = np.array(list(self._utilpairs[sid_index_out])) #!!!make more efficient with npfromiter?
        snpreader1 = self.snpreader1 if self.snpreader1 is not None else self.snpreader0
        sid_index_inner_0 = self.snpreader0.sid_to_index(pair_array[:,0])
        sid_index_inner_1 = snpreader1.sid_to_index(pair_array[:,1])


        if self.snpreader1 is None:
            sid_index_inner_01 = np.unique(np.r_[sid_index_inner_0,sid_index_inner_1]) #Index of every snp of interest
            inner_01 = self.snpreader0[iid_index_or_none,sid_index_inner_01].read(order=order, dtype=dtype, force_python_only=force_python_only, view_ok=True, num_threads=num_threads) #read every val of interest
            val_inner_01 = inner_01.standardize().val if self.do_standardize else inner_01.val

            sid_index_inner_01_reverse = {v:i for i,v in enumerate(sid_index_inner_01)} #Dictionary of snp_index to position in sid_index_inner_01
            sid_index_inner_0_in_val = np.array([sid_index_inner_01_reverse[i] for i in sid_index_inner_0])  #Replace snp_index0 with column # in val_inner_01
            sid_index_inner_1_in_val = np.array([sid_index_inner_01_reverse[i] for i in sid_index_inner_1])  #Replace snp_index1 with column # in val_inner_01
            val_inner_0 = val_inner_01[:,sid_index_inner_0_in_val] #Extract the vals for the left snps of interest
            val_inner_1 = val_inner_01[:,sid_index_inner_1_in_val]#Extract the vals for the right snps of interest
        else:
            inner_0 = self.snpreader0[iid_index_or_none,sid_index_inner_0].read(order=order, dtype=dtype, force_python_only=force_python_only, view_ok=True, num_threads=num_threads) #read every val of interest
            inner_1 = self.snpreader1[iid_index_or_none,sid_index_inner_1].read(order=order, dtype=dtype, force_python_only=force_python_only, view_ok=True, num_threads=num_threads) #read every val of interest
            val_inner_0 = inner_0.standardize().val if self.do_standardize else inner_0.val
            val_inner_1 = inner_1.standardize().val if self.do_standardize else inner_1.val
        val = val_inner_0*val_inner_1 #Element multiplication creates the vals for the pairs
        return val

#!!! keep these?

def split_on_sids(snpreader,part_count):
    sid_count = snpreader.sid_count
    start = 0
    for part_index in range(1,part_count+1):
        end = part_index*sid_count//part_count
        yield snpreader[:,start:end]
        start=end

def epi_reml(pair_snps,pheno,covar=None,kernel_snps=None,output_dir='results',part_count=33,runner=None,override=False):
    from pysnptools.kernelreader import SnpKernel
    from pysnptools.standardizer import Unit
    import datetime
    from fastlmm.association import single_snp

    part_list = list(split_on_sids(pair_snps,part_count))
    part_pair_count = (part_count*part_count+part_count)/2
    part_pair_index = -1
    print("part_pair_count={0:,}".format(part_pair_count))
    K0 = SnpKernel(kernel_snps or pair_snps,standardizer=Unit()).read() #Precompute the similarity
    if not os.path.exists(output_dir): os.makedirs(output_dir)
    start_time = datetime.datetime.now()
    for i in range(part_count):
        part_i = part_list[i]        
        for j in range(i,part_count):
            part_pair_index+=1
            pairs = _Pairs2(part_i) if i==j else _Pairs2(part_i,part_list[j])
            print("Looking at pair {0},{1} which is {2} of {3}".format(i,j,part_pair_index,part_pair_count))
            output_file = '{0}/result.{1}.{2}.tsv'.format(output_dir,part_pair_index,part_pair_count)
            if override or not os.path.exists(output_file):
                result_df_ij = single_snp(pairs, K0=K0, pheno=pheno, covar=covar, leave_out_one_chrom=False, count_A1=True,runner=runner)
                result_df_ij.to_csv(output_file, sep="\t", index=False)
                print(result_df_ij[:1])
                time_so_far = datetime.datetime.now()-start_time
                total_time_estimate = time_so_far*part_pair_count/(part_pair_index+1)
                print(total_time_estimate)


    
class TestPairs(unittest.TestCase):

    def test_run1(self):
        from pysnptools.snpreader import Bed
        root = Path(r'D:\OneDrive\programs\epireml') #!!! make this work without this
        runner = None
        #runner = LocalMultiProc(multiprocessing.cpu_count(),just_one_process=False)

        bed_original = Bed(str(root / 'syndata.bed'),count_A1=False)        
        pheno= str(root/'pheno.txt')
        covar= str(root/'cov.txt') #set to None, if none
        output_dir = r'm:/deldir/results.run1'
        part_count = 2

        epi_reml(bed_original[:,-20:],pheno,covar=covar,output_dir=output_dir,part_count=part_count,runner=runner,override=True)        
        #!!! check answer?

    def test_run2(self):
        from pysnptools.snpreader import Bed
        root = Path(r'D:\OneDrive\programs\epireml') #!!! make this work without this
        runner = None
        #runner = LocalMultiProc(multiprocessing.cpu_count(),just_one_process=False)
        bed_original = Bed(str(root / 'syndata.bed'),count_A1=False) #Read only the first 10 SNPs
        pheno= str(root / 'pheno.txt')
        covar= str(root / 'cov.txt')
        output_dir = r'm:/deldir/results.run2'
        part_count = 1

        epi_reml(bed_original[:,:20],pheno,kernel_snps=bed_original,covar=covar,output_dir=output_dir,part_count=part_count,runner=runner,override=True)

        #!!! check answer? against M:\deldir\refresults.run1 and 2

def getTestSuite():
    """
    set up composite test suite
    """
    
    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPairs))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    

    TestPairs().test_run1()
    TestPairs().test_run2()

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    r.run(suites)

    result = doctest.testmod()
    assert result.failed == 0, "failed doc test: " + __file__


    print("done")


    from pysnptools.snpreader import Pheno, Bed
    import pysnptools.util as pstutil

    if False:
        from pysnptools.snpreader import Bed
        from pysnptools.util import example_file # Download and return local file name
        bed_file = example_file('doc/ipynb/all.*','*.bed')
        bed = Bed(bed_file,count_A1=False)
        print(f'{bed.sid_count:,}')
        pairs = _Pairs(bed)
        print(f'{pairs.sid_count:,}')
        subset = pairs[:10,10000:10010]
        print(subset.read().val.shape)
        print(subset.sid)

    if False:

        data_file = 'd:\OneDrive\programs\epiCornell\syndata.bed'
        if False:
            from pysnptools.snpreader import SnpData
            import numpy as np
            bed1 = Bed("../../tests/datasets/synth/all")
            print(bed1.iid_count, bed1.sid_count, bed1.iid_count * bed1.sid_count)
            #goal 1500 individuals x 27000 SNP
            snpdata1 = bed1.read()
            iid = bed1.iid
            sid = ['sid{0}'.format(i) for i in xrange(27000)]
            val = np.tile(snpdata1.val,(3,6))[:,:27000].copy()
            #snpdata = Pheno('pysnptools/examples/toydata.phe').read()         # Read data from Pheno format
            snpdata2 = SnpData(iid, sid, val)
            print(snpdata2.iid_count, snpdata2.sid_count, snpdata2.iid_count * snpdata2.sid_count)
            Bed.write(snpdata2,data_file,count_A1=False)

        synbed = Bed(data_file)
        print(synbed.iid_count, synbed.sid_count, synbed.iid_count * synbed.sid_count)

        part_count = 1000
        part_list = list(split_on_sids(synbed,part_count))

        pairs00 = _Pairs(part_list[0])
        from fastlmm.association import single_snp
        pheno_fn = r"d:\OneDrive\programs\epiCornell\pheno.txt"
        cov_fn = r"d:\OneDrive\programs\epiCornell\cov.txt"
        results_df = single_snp(pairs00, K0=synbed, pheno=pheno_fn, covar=cov_fn, leave_out_one_chrom=False, count_A1=True)

        if False:
            for i,synbed_part_i in enumerate(synbed_part_list):
                for j,synbed_part_j in enumerate(synbed_part_list):
                    if j<i:
                        continue #not break
                    print("Looking at pair {0},{1}".format(i,j))
                    pairs = _Pairs(synbed_part_i) if i==j else _Pairs(synbed_part_i,synbed_part_j)
                    #print(pairs.iid)
                    print('{:,}'.format(pairs.sid_count))
                    #print(pairs.sid)
                    #print(pairs.pos)
                    #print(pairs.row_property)
                    snpdata = pairs.read()#
                    #print(snpdata.val)

        import datetime
        from pysnptools.kernelreader import SnpKernel
        from pysnptools.standardizer import Unit
        from pysnptools.util.mapreduce1.runner import LocalMultiProc
        from pysnptools.util.mapreduce1 import map_reduce
        #runner=None
        runner = LocalMultiProc(1,just_one_process=False)

        part_pair_count = (part_count*part_count+part_count)//2
        part_pair_index = -1
        print("part_pair_count={0:,}".format(part_pair_count))

        K0 = SnpKernel(synbed,standardizer=Unit()).read() #Precompute the similarity

        start_time = datetime.datetime.now()
        for i,part_i in enumerate(part_list):
            def mapper1(j):
                #from fastlmm.association import single_snp
                #from pysnptools.snpreader import Pairs
                #print('Z')
                #part_j = part_list[j]
                #print('A')
                print("Looking at pair {0},{1} which is {2} of {3}".format(i,j,part_pair_index+j+1,part_pair_count))
                #pairs = Pairs(part_i) if i==j else Pairs(part_i,part_j)
                #result_df_ij = single_snp(pairs, K0=K0, pheno=pheno_fn, covar=cov_fn, leave_out_one_chrom=False, count_A1=True)
                #print(result_df_ij[:1])
                #return result_df_ij

            result_df_i = map_reduce(range(i,part_count),
                                     mapper=mapper1,
                                     reducer=lambda result_j_list:pd.concat(result_j_list),
                                     runner=runner,
                                     name='js')
            part_pair_index+=(part_count-i)
            time_so_far = datetime.datetime.now()-start_time
            total_time_estimate = time_so_far*part_pair_count/(part_pair_index+1)
            print(total_time_estimate)




import numpy as np
from itertools import chain,islice
from  more_itertools import unique_everseen
import doctest
import unittest
import logging
import numbers
import pysnptools.util as pstutil

class _Pairs(object):
    """!!!description of class"""

    def __init__(self, list0, list1, include_singles, duplicates_ok=True): #!!! rename list to iterable?
        self.include_singles = include_singles
        self.duplicates_ok = duplicates_ok
        self.list0 = list(unique_everseen(list0))
        self.list1 = list(unique_everseen(list1))
        if not duplicates_ok:
            assert len(list0)==len(self.list0) and len(list1)==len(self.list1), "Expect lists to have no duplicates"

        dict0 = {k: v for v, k in enumerate(self.list0)}
        dict1 = {k: v for v, k in enumerate(self.list1)}
        self._common = sorted(six.viewkeys(dict0) & self.list1, key=lambda k:dict0[k])
        self._only0 = sorted(six.viewkeys(dict0)-self._common,key=lambda k:dict0[k])
        self._only1 = sorted(six.viewkeys(dict1)-self._common,key=lambda k:dict1[k])
    
        #How many pairs?
        self._count0 = len(self._only0)*len(self.list1)
        self._count1 = self._count1_fun(len(self._common),len(self._common),len(self._only1),include_singles)
        self._count = self._count0+self._count1

    def __len__(self):
        return self._count

    def __getitem__(self, indexer,speed='fast'):
        if isinstance(indexer,slice):
            assert indexer.step is None, "step is not supported"
            start,stop = indexer.start,indexer.stop
            if start is None:
                start = 0
            if stop is None:
                stop = self._count
            if start < 0:
                start = max(0,self._count+start)
            if stop < 0:
                stop = self._count+stop
            for pair in self._pair_sequence(start,stop,speed=speed):
                yield pair
        elif isinstance(indexer, numbers.Integral):
            start = indexer
            if start < 0:
                start = self._count+start
            assert 0<=start and start<self._count, 'Expect index between 0 (inclusive) and count of pairs (exclusive)'
            stop = start+1
            for pair in self._pair_sequence(start,stop,speed=speed):
                yield pair
        elif isinstance(indexer,np.ndarray): #!!! what about other arrays? iterator? negative elements?
            #!!! need test for this
            iterator = None
            previous = None
            for i in indexer:
                if iterator is None or i != previous+1:
                    iterator = self._pair_sequence(i,None,speed=speed)
                yield next(iterator)
                previous = i
        else:
            raise Exception("Don't understand indexer '{0}'".format(indexer))


    def _pair_sequence(self, start=0,stop=None,speed='fast'):
        start = int(start)
        if speed=='fast':
            stop = min(stop,self._count) if stop is not None else self._count
            return islice(self._pair_sequence_inner(start=start),max(stop-start,0))
        if speed=='medium':
            stop = stop if stop is not None else self._count
            return islice(self._pair_sequence_inner_medium(start=start),max(stop-start,0))
        else:
            assert speed=='slow',"Don't know speed '{0}'".format(speed)
            return islice(self._pair_sequence_inner_slow(),start,stop)

    def _pair_sequence_inner(self,start=0):
        if start >= self._count0:
            start -= self._count0
        else:
            row_start0 = start // len(self.list1)
            start -= row_start0*len(self.list1)
            assert 0 <= start and start <= len(self.list1), "real assert"
            for v0 in self._only0[row_start0:]:
                for v1 in islice(chain(self._only1,self._common),start,None):
                    yield v0,v1
                start = 0
        
        row_start1 = self._row_start1_fun(start,len(self._common),len(self._only1),self.include_singles)
        start -= self._count1_fun(row_start1,len(self._common),len(self._only1),self.include_singles)
        assert 0 <= start and start <= len(self.list1), "real assert"
        for common_index in range(row_start1,len(self._common)):
            v0 = self._common[common_index]
            common_start = common_index if self.include_singles else common_index+1
            assert start < len(self.list1)-common_start, "real assert"
            for v1 in islice(chain(self._only1,self._common[common_start:]),start,None):
                yield v0,v1
            start = 0

    def _pair_sequence_inner_medium(self,start=0):
        assert 0 <= start, "real assert"
        for v0 in self._only0:
            if start > len(self.list1):
                start -= len(self.list1)
            else:
                for v1 in islice(chain(self._only1,self._common),start,None):
                    yield v0,v1
                start = 0

        assert 0 <= start, "real assert"
        for index in range(len(self._common)):
            v0 = self._common[index]
            startx = index if self.include_singles else index+1
            if start > len(self.list1)-startx:
                start -= (len(self.list1)-startx)
            else:
                for v1 in islice(chain(self._only1,self._common[startx:]),start,None):
                    yield v0,v1
                start = 0

    def _pair_sequence_inner_slow(self):

        for v0 in self._only0:
            for v1 in chain(self._only1,self._common):
                yield v0,v1

        for index in range(len(self._common)):
            v0 = self._common[index]
            startx = index if self.include_singles else index+1
            for v1 in chain(self._only1,self._common[startx:]):
                yield v0,v1


    @staticmethod
    def _count1_fun(row_start,len_common,len_only1,include_singles):
        a2 = -1
        b2 = len_only1*2+len_common*2+(1 if include_singles else -1)
        count1 = (a2*row_start*row_start + b2*row_start)//2
        return count1

    @staticmethod
    def _row_start1_fun(count1,len_common,len_only1,include_singles):
        a2 = -1
        b2 = len_only1*2+len_common*2+(1 if include_singles else -1)
        c = -count1
        row_start = int((b2-(b2*b2-8*a2*c)**.5)/2.0) #!!! will this ever be off by one because of numerical problems?
        if row_start < 0:
            row_start = int((b2+(b2*b2-8*a2*c)**.5)/2.0)
        return row_start

class TestPairs(unittest.TestCase):

    def test_count1_fun(self):
        for len_common in range(5):
            for len_only1 in range(5):
                for include_singles in [True,False]:
                    list0 = ['common{0}'.format(i) for i in range(len_common)]
                    list1 = list0 + ['only1_{0}'.format(i) for i in range(len_only1)]
                    pairs = _Pairs(list0,list1,include_singles)
                    pair_list = list(pairs._pair_sequence(0,speed='slow'))
                    for row_start in range(len_common+1):
                        count = _Pairs._count1_fun(row_start,len_common,len_only1,include_singles)
                        count_ref = len([pair for pair in pair_list if pair[0] in list0[:row_start]])
                        assert count==count_ref, "_count1_fun isn't giving the right answer"

    def test_row_start_fun(self):
        for len_common in range(5):
            for len_only1 in range(5):
                for include_singles in [True,False]:
                    list0 = ['common{0}'.format(i) for i in range(len_common)]
                    list1 = list0 + ['only1_{0}'.format(i) for i in range(len_only1)]
                    pairs = _Pairs(list0,list1,include_singles)
                    pair_list = list(pairs._pair_sequence(0,speed='slow'))
                    for start in range(len(pairs)+1):
                        row_start = pairs._row_start1_fun(start,len_common,len_only1,include_singles)
                        assert start==len(pair_list) or pair_list[start][0]==list0[row_start], "_row_start1_fun isn't giving the right answer"

    def test_index_and_count1_functions(self):
        for len_commonq in range(0,5):  
            for len_only1q in range(0,5):
                for include_singlesq in [True,False]:
                    for row_startq in range(len_commonq+1):
                        count1q =_Pairs._count1_fun(row_startq,len_commonq,len_only1q,include_singlesq)
                        row_start2 = _Pairs._row_start1_fun(count1q,len_commonq,len_only1q,include_singlesq)
                        count2 = _Pairs._count1_fun(row_start2,len_commonq,len_only1q,include_singlesq)
                        assert count1q==count2

    def test_medium_fast(self):
        for len_common in range(2):
            common = ['common{0}'.format(i) for i in range(len_common)]
            for len_only0 in range(2):
                list0 = common + ['only0_{0}'.format(i) for i in range(len_only0)]
                for len_only1 in range(2):
                    list1 = common + ['only1_{0}'.format(i) for i in range(len_only1)]
                    for include_singles in [True,False]:
                        pairs = _Pairs(list0, list1, include_singles, duplicates_ok=True)
                        for goal_see in range(len(pairs)+1):
                            for start in range(len(pairs)+1):
                                logging.info((len_common,len_only0,len_only1,include_singles,start,start+goal_see))
                                slow = np.array(list(pairs._pair_sequence(start,start+goal_see,speed='slow')))
                                medium = np.array(list(pairs._pair_sequence(start,start+goal_see,speed='medium')))
                                assert np.array_equal(slow,medium), 'Expect slow and medium to give the same answers'
                                fast = np.array(list(pairs[start:start+goal_see]))
                                assert np.array_equal(medium,fast), 'Expect medium and fast to give the same answers'
    def test_big(self):
        size = 500*1000
        seed = 0
        np.random.seed(seed)
        list0 = np.random.randint(size*10,size=size)
        list1 = np.random.randint(size*10,size=size)
        for include_singles in [True,False]:
            pairs = _Pairs(list0, list1, include_singles, duplicates_ok=True)
            for start in [0,len(pairs)//5,len(pairs)-1,len(pairs)]:
                logging.info(("test_big",include_singles,start))
                medium = np.array(list(pairs._pair_sequence(start,start+10,speed='medium')))
                fast = np.array(list(pairs[start:start+10]))
                assert np.array_equal(medium,fast), 'Expect medium and fast to give the same answers'


    #!!! make doc string example with first and last names
    def test_slice(self):
        for len_common in range(3):
            common = ['common{0}'.format(i) for i in range(len_common)]
            for len_only0 in range(3):
                list0 = common + ['only0_{0}'.format(i) for i in range(len_only0)]
                for len_only1 in range(3):
                    list1 = common + ['only1_{0}'.format(i) for i in range(len_only1)]
                    for include_singles in [True,False]:
                        pairs = _Pairs(list0, list1, include_singles, duplicates_ok=True)
                        pair_list = np.array(list(pairs._pair_sequence(0,None,speed='slow')),dtype='U')
                        for slicer, start, stop in [
                            (-1,len(pairs)-1,len(pairs)),#[-1]
                            (slice(None,None),0,len(pairs)),#[:]
                            (slice(None,2),0,2),#[:2]
                            (2,2,3),#[2]
                            (slice(2,None),2,None),#[2:]
                            (slice(1,3),1,3),#[1:3]
                            (slice(3,3),3,3),#[3:3]
                            (slice(-3,-2),len(pairs)-3,len(pairs)-2),#[-3:-2]
                            (slice(-3,None),len(pairs)-3,None),#[-3:],
                            (slice(None,-3),None,len(pairs)-3),#[:-3],
                            (slice(1000,None),1000,None),#[1000:]
                            (slice(-3,None),len(pairs)-100000,None),#[-100000:]
                            (slice(4,1),4,1),#[4:1],
                            (slice(10,1000),10,1000)#[10,1000]
                            ]:
                            logging.info(("test_slice",slicer, start, stop,len_common,len_only0,len_only1,include_singles))
                            try:
                                slow = pair_list[slicer]
                                slow_fail = False
                            except:
                                slow_fail = True
                            try:
                                fast = np.array(list(pairs[slicer]),dtype='U')
                                fast_fail = False
                            except:
                                fast_fail = True
                            if slow_fail:
                                assert fast_fail, 'expect slow and fast to fail on the same inputs'
                            else:
                                if fast_fail or not np.array_equal(slow.reshape(-1,2),fast.reshape(-1,2)): #!!!:
                                    fast = np.array(list(pairs[slicer]))
                                assert not fast_fail, 'expect slow and fast to fail on the same inputs'
                                assert np.array_equal(slow.reshape(-1,2),fast.reshape(-1,2)), 'Expect slow and fast to give the same answers'

def getTestSuite():
    """
    set up composite test suite
    """
    
    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPairs))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    #!!! this is work on an example for docstring, but a bit weird because names are ordered and these arent
    pairs = _Pairs(['Madison','Jill','Kelsey'],['Smith','Madison','Kelsey'],include_singles=True)#!!! is 'include_singles' the best name?
    for first,last in pairs[:]:
        print(first,last)

    pairs = _Pairs(range(0,2000),range(1000,3000),include_singles=True)
    print(len(pairs))
    for pair in pairs[1999999:2000004]:
        print(pair)
    print(list(pairs[-5:]))


    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    r.run(suites)

    result = doctest.testmod()
    assert result.failed == 0, "failed doc test: " + __file__


    print("done")